#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.util.error import NotIncludedError
from covsirphy.util.alias import Alias
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.gis.gis import GIS
from covsirphy.downloading.downloader import DataDownloader
from covsirphy.engineering.cleaner import _DataCleaner
from covsirphy.engineering.transformer import _DataTransformer
from covsirphy.engineering.complement import _ComplementHandler


class DataEngineer(Term):
    """Class for data engineering including loading, cleaning, transforming, complementing, EDA (explanatory data analysis).

    Args:
        layers (list[str] or None): list of layers of geographic information or None (["ISO3", "Province", "City"])
        country (str or None): layer name of countries or None (countries are not included in the layers)
        verbose (int): level of verbosity of stdout

        Raises:
            ValueError: @layers has duplicates

        Note:
            Country level data specified with @country will be stored with ISO3 codes.

        Note:
            If @verbose is 0, no descriptions will be shown.
            If @verbose is 1 or larger, details of layer adjustment will be shown.
    """

    def __init__(self, layers=None, country="ISO3", verbose=1):
        self._layers = Validator(layers, "layers").sequence(default=[self.ISO3, self.PROVINCE, self.CITY])
        self._country = str(country)
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))
        self._gis_kwargs = dict(layers=self._layers, country=self._country, date=self.DATE, verbose=verbose)
        self._gis = GIS(**self._gis_kwargs)
        # Aliases
        self._var_alias = Alias.for_variables()
        self._subset_alias = Alias(target_class=tuple)

    def register(self, data, citations=None, **kwargs):
        """Register new data.

        Args:
            data (pandas.DataFrame): new data
                Index
                    reset index
                Columns
                    columns defined by covsirphy.DataEngineer(layer)
                    Date (pandas.DataFrame): observation dates
                    Population (int): total population, optional
                    Tests (int): column of the number of tests, optional
                    Confirmed (int): the number of confirmed cases, optional
                    Fatal (int): the number of fatal cases, optional
                    Recovered (int): the number of recovered cases, optional
                    the other columns will be also registered
            citations (list[str] or str or None): citations of the dataset or None (["my own dataset"])
            **kwargs: keyword arguments of pandas.to_datetime() including "dayfirst (bool): whether date format is DD/MM or not"

        Returns:
            covsirphy.DataEngineer: self
        """
        Validator(data, "data").dataframe(columns=[*self._layers, self.DATE])
        self._gis.register(
            data=data, layers=self._layers, date=self.DATE, variables=None,
            citations=citations or ["my own dataset"], convert_iso3=(self._country in self._layers), **kwargs)
        return self

    def download(self, **kwargs):
        """Download datasets from the recommended data servers using covsirphy.DataDownloader.

        Args:
            **kwargs: keyword arguments of covsirphy.DataDownloader() and covsirphy.DataDownloader.layer()

        Returns:
            covsirphy.DataEngineer: self

        Note:
            When @verbose is not included in **kwargs, covsirphy.DataEngineer(verbose) will be used.
        """
        default_dict = {"verbose": self._verbose}
        validator = Validator(kwargs, name="keyword arguments")
        downloader = DataDownloader(**validator.kwargs(DataDownloader, default=default_dict))
        df = downloader.layer(**validator.kwargs(DataDownloader.layer, default=default_dict))
        citations = downloader.citations()
        self._gis.register(
            data=df, layers=[self.ISO3, self.PROVINCE, self.CITY], date=self.DATE, variables=None,
            citations=citations, convert_iso3=False, **kwargs)
        return self

    def all(self, variables=None):
        """Return all available data, converting dtypes with pandas.DataFrame.convert_dtypes().

        Args:
            variables (list[str] or str or None): list of variables to collect or alias or None (all available variables)

        Raises:
            NotRegisteredError: No records have been registered yet

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Column
                    columns defined by @layers of _DataEngineer()
                    (pandas.Timestamp): observation dates defined by @date of _DataEngineer()
                    the other columns
        """
        return self._gis.all(variables=self._var_alias.find(name=variables, default=variables), errors="raise").convert_dtypes()

    def citations(self, variables=None):
        """
        Return citation list of the secondary data sources.

        Args:
            variables (list[str] or str or None): list of variables to collect or alias or None (all available variables)

        Returns:
            list[str]: citation list
        """
        return self._gis.citations(variables=self._var_alias.find(name=variables, default=variables))

    def clean(self, kinds=None, **kwargs):
        """Clean all registered data.

        Args:
            kinds (list[str] or None): kinds of data cleaning with order or None (all available kinds as follows)
                - "convert_date": Convert dtype of date column to pandas.Timestamp.
                - "resample": Resample records with dates.
                - "fillna": Fill NA values with '-' (layers) and the previous values and 0.
            **kwargs: keyword arguments of data cleaning refer to note

        Returns:
            covsirphy.DataEngineer: self

        Note:
            For "convert_date", keyword arguments of pandas.to_datetime() including "dayfirst (bool): whether date format is DD/MM or not" can be used.
        """
        citations = self._gis.citations(variables=None)
        cleaner = _DataCleaner(data=self._gis.all(), layers=self._layers, date=self.DATE)
        kind_dict = {
            "convert_date": cleaner.convert_date,
            "resample": cleaner.resample,
            "fillna": cleaner.fillna,
        }
        all_kinds = list(kind_dict.keys())
        selected = Validator(kinds, "kind").sequence(default=all_kinds, candidates=all_kinds)
        for kind in selected:
            kind_dict[kind](**Validator(kwargs, "keyword arguments").kwargs(functions=kind_dict[kind], default=None))
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=cleaner.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def transform(self):
        """Transform all registered data, calculating the number of susceptible and infected cases.

        Returns:
            covsirphy.DataEngineer: self

        Note:
            Susceptible = Population - Confirmed.

        Note:
            Infected = Confirmed - Fatal - Recovered.
        """
        all_df = self._gis.all()
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=all_df, layers=self._layers, date=self.DATE)
        transformer.susceptible(new=self.S, population=self.N, confirmed=self.C)
        transformer.infected(new=self.CI, confirmed=self.C, fatal=self.F, recovered=self.R)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def inverse_transform(self):
        """Perform inverse transformation, calculating total population and confirmed.

        Returns:
            covsirphy.DataEngineer: self

        Note:
            Population = Susceptible + Confirmed.

        Note:
            Confirmed = Infected + Fatal + Recovered.
        """
        Validator(self._gis.all(), "all registered data").dataframe(columns=[self.S, self.CI, self.F, self.R])
        self.add(columns=[self.CI, self.F, self.R], new=self.C)
        self.add(columns=[self.S, self.C], new=self.N)
        return self

    def diff(self, column, suffix="_diff", freq="D"):
        """Calculate daily new cases with "f(x>0) = F(x) - F(x-1), x(0) = 0 when F is cumulative numbers".

        Args:
            column (str): column name of the cumulative numbers
            suffix (str): suffix if the column (new column name will be '{column}{suffix}')
            freq (str): offset aliases of shifting dates

        Returns:
            covsirphy.DataEngineer: self

        Note:
            Regarding @freq, refer to https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        Note:
            If the alias of values
        """
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.diff(
            column=Validator(
                self._var_alias.find(name=column, default=[column]), "column", accept_none=False).sequence(length=1)[0],
            suffix=suffix, freq=freq)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def add(self, columns, new=None, fill_value=0):
        """Calculate element-wise addition with pandas.DataFrame.sum(axis=1), X1 + X2 + X3 +...

        Args:
            columns (str): columns to add
            new (str): column name of addition or None (f"{X1}+{X2}+{X3}...")
            fill_value (float): value to fill in NAs

        Returns:
            covsirphy.DataEngineer: self
        """
        col_names = self._var_alias.find(name=columns, default=columns)
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.add(columns=col_names, new=new or "+".join(col_names), fill_value=fill_value)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def mul(self, columns, new=None, fill_value=0):
        """Calculate element-wise multiplication with pandas.DataFrame.product(axis=1), X1 * X2 * X3 *...

        Args:
            columns (str): columns to multiply
            new (str): column name of multiplication or None (f"{X1}*{X2}*{X3}...")
            fill_value (float): value to fill in NAs

        Returns:
            covsirphy.DataEngineer: self
        """
        col_names = self._var_alias.find(name=columns, default=columns)
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.mul(columns=col_names, new=new or "*".join(col_names), fill_value=fill_value)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def sub(self, minuend, subtrahend, new=None, fill_value=0):
        """Calculate element-wise subtraction with pandas.Series.sub(), minuend - subtrahend.

        Args:
            minuend (str): numerator column
            subtrahend (str): subtrahend column
            new (str): column name of subtraction or None (f"{minuend}-{subtrahend}")
            fill_value (float): value to fill in NAs

        Returns:
            covsirphy.DataEngineer: self
        """
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.sub(
            minuend=Validator(
                self._var_alias.find(name=minuend, default=[minuend]), "minuend", accept_none=False).sequence(length=1)[0],
            subtrahend=Validator(
                self._var_alias.find(name=subtrahend, default=[subtrahend]), "subtrahend", accept_none=False).sequence(length=1)[0],
            new=new or f"{minuend}-{subtrahend}", fill_value=fill_value)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def div(self, numerator, denominator, new=None, fill_value=0):
        """Calculate element-wise floating division with pandas.Series.div(), numerator / denominator.

        Args:
            numerator (str): numerator column
            denominator (str): denominator column
            new (str or None): column name of floating division or None (f"{numerator}_per_({denominator.replace(' ', '_')})")
            fill_value (float): value to fill in NAs

        Returns:
            covsirphy.DataEngineer: self

        Note:
            Positive rate could be calculated with Confirmed / Tested, `.div(numerator="Confirmed", denominator="Tested", new="Positive_rate")`
        """
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.div(
            numerator=Validator(
                self._var_alias.find(name=numerator, default=[numerator]), "numerator", accept_none=False).sequence(length=1)[0],
            denominator=Validator(
                self._var_alias.find(name=denominator, default=[denominator]), "denominator", accept_none=False).sequence(length=1)[0],
            new=new or f"{numerator}_per_({denominator.replace(' ', '_')})", fill_value=fill_value)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def assign(self, **kwargs):
        """Assign a new column with pandas.DataFrame.assign().

        Args:
            **kwargs: dict of {str: callable or pandas.Series}

        Note:
            Refer to documentation of pandas.DataFrame.assign(), https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.assign.html
        """
        citations = self._gis.citations(variables=None)
        transformer = _DataTransformer(data=self._gis.all(), layers=self._layers, date=self.DATE)
        transformer.assign(**kwargs)
        self._gis = GIS(**self._gis_kwargs)
        self._gis.register(
            data=transformer.all(), layers=self._layers, date=self.DATE, variables=None, citations=citations, convert_iso3=False)
        return self

    def layer(self, geo=None, start_date=None, end_date=None, variables=None):
        """Return the data at the selected layer in the date range.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to specify the layer or None (the top level)
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            variables (list[str] or None): list of variables to add or None (all available columns)

        Raises:
            TypeError: @geo has un-expected types
            ValueError: the length of @geo is larger than the length of layers
            NotRegisteredError: No records have been registered at the layer yet

        Returns:
            pandas.DataFrame:
                Index:
                    reset index
                Columns
                    - (str): columns defined by covsirphy.GIS(layers)
                    - Date (pandas.Timestamp): observation dates
                    - columns defined by @variables

        Note:
           Note that records with NAs as country names will be always removed.

        Note:
            Regarding @geo argument, please refer to covsirphy.GIS.layer().
        """
        v_converted = self._var_alias.find(name=variables, default=variables)
        return self._gis.layer(geo=geo, start_date=start_date, end_date=end_date, variables=v_converted, errors="raise")

    def choropleth(self, geo, variable, on=None, title="Choropleth map", filename="choropleth.jpg", logscale=True, directory=None, natural_earth=None, **kwargs):
        """Create choropleth map.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to specify the layer or None (the top level)
            variable (str): variable name to show
            on (str or None): the date, like 22Jan2020, or None (the last date of each location)
            title (str): title of the map
            filename (str or None): filename to save the figure or None (display)
            logscale (bool): whether convert the value to log10 scale values or not
            directory (str or None): directory to save GeoJSON file of "Natural Earth" GitHub repository or None (the directory of GIS class script)
            natural_earth (str or None): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository or None (automatically determined)
            **kwargs: keyword arguments of the following classes and methods.
                - matplotlib.pyplot.savefig(), matplotlib.pyplot.legend(), and
                - pandas.DataFrame.plot()

        Note:
            Regarding @geo argument, please refer to covsirphy.GIS.layer().

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
            https://www.naturalearthdata.com/
            https://github.com/nvkelso/natural-earth-vector
            Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)
        """
        layer_df = self.layer(geo=geo, variables=[variable])
        gis = GIS(**self._gis_kwargs)
        gis.register(data=layer_df, date=self.DATE)
        gis.choropleth(
            variable=variable, filename=filename, title=title, logscale=logscale,
            geo=geo, on=on, directory=directory, natural_earth=natural_earth, **kwargs)

    def subset(self, geo=None, start_date=None, end_date=None, variables=None, complement=True, **kwargs):
        """Return subset of the location and date range.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to filter or None (total at the top level)
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            variables (list[str] or None): list of variables to add or None (all available columns)
            complement (bool): whether perform data complement or not, True as default
            **Kwargs: keyword arguments for complement and default values
                recovery_period (int): expected value of recovery period [days], 17
                interval (int): expected update interval of the number of recovered cases [days], 2
                max_ignored (int): Max number of recovered cases to be ignored [cases], 100
                max_ending_unupdated (int): Max number of days to apply full complement, where max recovered cases are not updated [days], 14
                upper_limit_days (int): maximum number of valid partial recovery periods [days], 90
                lower_limit_days (int): minimum number of valid partial recovery periods [days], 7
                upper_percentage (float): fraction of partial recovery periods with value greater than upper_limit_days, 0.5
                lower_percentage (float): fraction of partial recovery periods with value less than lower_limit_days, 0.5

        Returns:
            tuple(pandas.DataFrame, str, dict):
                pandas.DataFrame
                    Index
                        Date (pandas.DataFrame): observation dates
                    Columns
                        Population (int): total population
                        Tests (int): column of the number of tests
                        Confirmed (int): the number of confirmed cases
                        Fatal (int): the number of fatal cases
                        Recovered (int): the number of recovered cases
                        the other columns registered
                str: status code: will be selected from
                    - '' (not complemented)
                    - 'monotonic increasing complemented confirmed data'
                    - 'monotonic increasing complemented fatal data'
                    - 'monotonic increasing complemented recovered data'
                    - 'fully complemented recovered data'
                    - 'partially complemented recovered data'
                dict[str, bool]: status for each complement type, keys are
                    - Monotonic_confirmed
                    - Monotonic_fatal
                    - Monotonic_recovered
                    - Full_recovered
                    - Partial_recovered

        Note:
            Regarding @geo argument, please refer to covsirphy.GIS.subset().

        Note:
            Re-calculation of Susceptible and Infected will be done automatically.
        """
        v_converted = self._var_alias.find(name=variables, default=variables)
        subset_df = self._gis.subset(geo=geo, start_date=start_date, end_date=end_date, variables=None, errors="raise")
        if not complement:
            df = subset_df.set_index(self.DATE)
            return df.loc[:, v_converted or df.columns].convert_dtypes(), "", {}
        default_kwargs = {
            "recovery_period": 17,
            "interval": 2,
            "max_ignored": 100,
            "max_ending_unupdated": 14,
            "upper_limit_days": 90,
            "lower_limit_days": 7,
            "upper_percentage": 0.5,
            "lower_percentage": 0.5,
        }
        handler = _ComplementHandler(
            **Validator(kwargs, "keyword arguments").kwargs(_ComplementHandler, default=default_kwargs))
        c_df, status, status_dict = handler.run(data=subset_df)
        df = pd.concat([subset_df.drop([self.DATE, self.C, self.F, self.R], axis=1), c_df], axis=1)
        df["location"] = self.NA
        transformer = _DataTransformer(data=df, layers=["location"], date=self.DATE)
        transformer.susceptible(new=self.S, population=self.N, confirmed=self.C)
        transformer.infected(new=self.CI, confirmed=self.C, fatal=self.F, recovered=self.R)
        transformed_df = transformer.all().drop("location", axis=1).set_index(self.DATE)
        return transformed_df.loc[:, v_converted or transformed_df.columns], status, status_dict

    def subset_alias(self, alias=None, update=False, **kwargs):
        """Set/get/list-up alias name(s) of subset.

        Args:
            alias (str or None): alias name or None (list-up alias names)
            update (bool): force updating the alias when @alias is not None
            **kwargs: keyword arguments of covsirphy.DataEngineer().subset()

        Returns:
            tuple(pandas.DataFrame, str, dict) or dict[str, tuple(pandas.DataFrame, str, dict)]:
                - tuple(pandas.DataFrame, str, dict): when @alias is not None, the subset of the alias
                - dict[str, tuple(pandas.DataFrame, str, dict)]: when @alias is None, dictionary of aliases and subsets

        Note:
            When the alias name was a new one, subset will be registered with covsirphy.DataEngineer.subset(**kwargs).
        """
        if alias is None:
            return self._subset_alias.all()
        result = self._subset_alias.find(alias, default=None)
        if update or result is None:
            self._subset_alias.update(name=alias, target=self.subset(**kwargs))
        return self._subset_alias.find(alias)

    def variables_alias(self, alias=None, variables=None):
        """Set/get/list-up alias name(s) of variables.

        Args:
            alias (str or None): alias name or None (list-up alias names)
            variables (list[str]): variables to register with the alias

        Raises:
            NotIncludedError: the alias is not None and un-registered

        Returns:
            list[str] or dict[str, list[str]]:
                - list[str]: when @alias is not None, the variables of the alias
                - dict[str, list[str]]: when @alias is None, dictionary of aliases and variables

        Note:
            When @variables is not None, alias will be registered/updated.

        Note:
            Some aliases are preset. We can check them with covsirphy.DataEngineer().variables_alias().
        """
        if alias is None:
            return self._var_alias.all()
        if variables is not None:
            self._var_alias.update(name=alias, target=variables)
        elif alias not in self._var_alias.all():
            raise NotIncludedError(alias, "keys of alias dictionary of variables")
        return self._var_alias.find(name=alias)

    @classmethod
    def recovery_period(cls, data):
        """Calculate mode value of recovery period of the data.

        Args:
            data (pandas.DataFrame): data for calculation
                Index
                    Date (pandas.Timestamp): observation dates
                Columns
                    Confirmed (int): the number of confirmed cases, optional
                    Fatal (int): the number of fatal cases, optional
                    Recovered (int): the number of recovered cases, optional
                    the other columns will be ignored

        Returns:
            int: mode value of recovery period [days]
        """
        df = Validator(data, "data").dataframe(time_index=True, columns=[cls.C, cls.F, cls.R], empty_ok=False)
        df = df.resample("D").sum()
        df["diff"] = df[cls.C] - df[cls.F]
        df = df.loc[:, ["diff", cls.R]].unstack().reset_index()
        df.columns = ["Variable", "Date", "Number"]
        df["Days"] = (df["Date"] - df["Date"].min()).dt.days
        df = df.pivot_table(values="Days", index="Number", columns="Variable")
        df = df.interpolate(limit_area="inside").dropna().astype(np.int64)
        df["Elapsed"] = df[cls.R] - df["diff"]
        df = df.loc[df["Elapsed"] > 0]
        return 0 if df.empty else round(df["Elapsed"].mode().mean())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from sklearn.model_selection import train_test_split
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _FeatureEngineer(Term):
    """
    Perform feature engineering and split X, Y to train/test/target datasets.

    Args:
        X (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float): indicators
        Y (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                (int/float) target values
    """

    def __init__(self, X, Y):
        # Validate values
        self._X_raw = Validator(X, "X").dataframe(time_index=True)
        self._X = X.copy()
        self._Y = Validator(Y, "Y").dataframe(time_index=True)

    def split(self, **kwargs):
        """
        Split the current X and Y to train/test/target datasets.

        Args:
            kwargs: keyword arguments of sklearn.model_selection.train_test_split()

        Returns:
            dict[str, pandas.DataFrame]: datasets with time index
                - X_train
                - X_test
                - Y_train
                - Y_test
                - X_target

        Note:
            If @seed is included in kwargs, this will be converted to @random_state.

        Note:
            default values regarding sklearn.model_selection.train_test_split() are
            test_size=0.2, random_state=0, shuffle=False.
        """
        split_kwargs = {"test_size": 0.2, "random_state": 0, "shuffle": False}
        split_kwargs.update(kwargs)
        split_kwargs["random_state"] = split_kwargs.get("seed", split_kwargs["random_state"])
        split_kwargs = Validator(split_kwargs, "keyword arguments").kwargs(functions=train_test_split)
        # Train/test
        X, Y = self._X.copy(), self._Y.copy()
        df = X.join(Y, how="inner").dropna().drop_duplicates()
        X_arranged = df.loc[:, X.columns]
        Y_arranged = df.loc[:, Y.columns]
        splitted = train_test_split(X_arranged, Y_arranged, **split_kwargs)
        names = ["X_train", "X_test", "Y_train", "Y_test"]
        data_dict = dict(zip(names, splitted))
        # X_target
        data_dict["X_target"] = X.loc[X.index > Y.index.max()]
        return data_dict

    def add_elapsed(self):
        """
        Calculate elapsed days from the last change point of indicators and add them as new features.
        """
        X, raw = self._X.copy(), self._X_raw.copy()
        for (name, col) in raw.iteritems():
            X[f"{name}_elapsed"] = col.groupby((col != col.shift()).cumsum()).cumcount() + 1
        self._X = X.copy()

    def log_transform(self):
        """
        Add log-transformed indicator values to X as new features.
        """
        raw = self._X_raw.copy()
        selected = raw.loc[:, raw.max(axis=0) >= 1000]
        converted = np.log10(selected + 1)
        self._X = self._X.join(converted, rsuffix="_log10")

    def apply_delay(self, delay_values):
        """
        Add delayed indicator values to X.

        Args:
            delay_values (list[int]): list of delay period [days]

        Note:
            This will be applied to all indicators, including features created by the other tools.
        """
        X = self._X.copy()
        X_copy = X.copy()
        for delay in delay_values:
            X = X.join(X_copy.shift(delay, freq="D"), how="outer", rsuffix=f"_{delay}")
        self._X = X.ffill()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from lightgbm import LGBMRegressor
from sklearn.decomposition import PCA
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit
from sklearn.multioutput import MultiOutputRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from covsirphy._deprecated.regbase import _RegressorBase
from covsirphy._deprecated.reg_rate_converter import _RateConverter


class _ParamLightGBMRegressor(_RegressorBase):
    """
    Predict parameter values of ODE models with light gradient boosting machine regressor.

    Args:
        - X_train (pandas.DataFrame): X for training with time index
        - X_test (pandas.DataFrame): X for test with time index
        - Y_train (pandas.DataFrame): Y for training with time index
        - Y_test (pandas.DataFrame): Y for test with time index
        - X_target (pandas.DataFrame): X for prediction with time index
    """
    # Description of regressor
    DESC = "Indicators -> Parameters with Light Gradient Boosting Machine Regressor"

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        """
        # Paramters of the steps
        param_grid = {
            "converter__to_convert": [True, False],
            "pca__n_components": [0.3, 0.5, 0.7, 0.9],
            "regressor__estimator__max_depth": list(range(1, 5)),
        }
        # Fit with pipeline
        steps = [
            ("converter", _RateConverter()),
            ("scaler", MinMaxScaler()),
            ("pca", PCA(random_state=0)),
            ("regressor", MultiOutputRegressor(LGBMRegressor(n_estimators=200, random_state=0))),
        ]
        tscv = TimeSeriesSplit(n_splits=5).split(self._X_train)
        pipeline = GridSearchCV(Pipeline(steps=steps), param_grid, n_jobs=-1, cv=tscv)
        pipeline.fit(self._X_train, self._Y_train)
        # Update regressor
        self._pipeline = pipeline
        # Update param
        self._param.update(**{k: type(v) for (k, v) in steps})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from autots import AutoTS
from sklearn.exceptions import ConvergenceWarning
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _AutoTSPredictor(Term):
    """
    Predict ODE parameter values with AutoTS library.

    Args:
        model_list (list[str]): list of machine learning models
        days (int): days to predict
        kwargs: keyword arguments of autots.AutoTS.

    Note:
        Candidates of @model_list can be checked with `from autots.models.model_list import model_lists; print(model_lists)`.
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=ConvergenceWarning)

    def __init__(self, days, model_list, **kwargs):
        autots_kwargs = {
            "forecast_length": days,
            "model_list": model_list,
            "frequency": "D",
            "ensemble": "horizontal",
            "max_generations": 1,
            "num_validations": 2,
            "validation_method": "backwards",
            "no_negatives": True,
            "constraint": 2.0,
            "random_seed": 0,
            "verbose": 1,
            "n_jobs": -1,
        }
        self._autots_kwargs = Validator(kwargs, "keyword arguments").kwargs(functions=AutoTS, default=autots_kwargs)

    def predict(self, past_df):
        """
        Run AutoTS prediction.

        Args:
            past_df (pandas.DataFrame): dataset for training/validation
                Index
                    pandas.Timestamp: Observation date
                Columns
                    observed variables (int or float)

        Returns:
            tuple(pandas.DataFrame, pandas.DataFrame, pandas.DataFrame): the most likely, upper, lower values
                Index
                    reset index
                Columns
                    predicted values (float)
        """
        model = AutoTS(**self._autots_kwargs)
        model = model.fit(past_df)
        prediction = model.predict()
        return (prediction.forecast, prediction.upper_forecast, prediction.lower_forecast)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.error import NotRegisteredError, SubsetNotFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.gis.gis import GIS
from covsirphy.downloading.db_cs_japan import _CSJapan
from covsirphy.downloading.db_covid19dh import _COVID19dh
from covsirphy.downloading.db_google import _GoogleOpenData
from covsirphy.downloading.db_owid import _OWID


class DataDownloader(Term):
    """Class to download datasets from the recommended data servers.

    Args:
        directory (str or pathlib.Path): directory to save downloaded datasets
        update_interval (int): update interval of downloading dataset
        verbose (int): level of verbosity when downloading

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, URL and database name will be shown.

    Note:
        Location layers are fixed to ["ISO3", "Province", "City"]
    """
    LAYERS = [Term.ISO3, Term.PROVINCE, Term.CITY]

    def __init__(self, directory="input", update_interval=12, verbose=1):
        self._directory = directory
        self._update_interval = Validator(update_interval, "update_interval").int(value_range=(0, None))
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))
        self._gis = GIS(layers=self.LAYERS, country=self.ISO3, date=self.DATE, verbose=self._verbose)

    def layer(self, country=None, province=None, databases=None):
        """Return the data at the selected layer.

        Args:
            country (str or None): country name or None
            province (str or None): province/state/prefecture name or None
            databases (list[str] or None): list of databases to use or None (all available databases).
                "japan": COVID-19 Dataset in Japan,
                "covid19dh": COVID-19 Data Hub,
                "google: COVID-19 Open Data by Google Cloud Platform,
                "owid": Our World In Data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Country (str): country names (top level administration)
                    - Province (str): province names (2nd level administration)
                    - ISO3 (str): ISO3 codes
                    - Confirmed (pandas.Int64): the number of confirmed cases
                    - Fatal (pandas.Int64): the number of fatal cases
                    - Recovered (pandas.Int64): the number of recovered cases
                    - Population (pandas.Int64): population values
                    - Tests (pandas.Int64): the number of tests
                    - Product (pandas.Int64): vaccine product names
                    - Vaccinations (pandas.Int64): cumulative number of vaccinations
                    - Vaccinations_boosters (pandas.Int64): cumulative number of booster vaccinations
                    - Vaccinated_once (pandas.Int64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (pandas.Int64): cumulative number of people who received all doses prescribed by the protocol
                    - School_closing
                    - Workplace_closing
                    - Cancel_events
                    - Gatherings_restrictions
                    - Transport_closing
                    - Stay_home_restrictions
                    - Internal_movement_restrictions
                    - International_movement_restrictions
                    - Information_campaigns
                    - Testing_policy
                    - Contact_tracing
                    - Stringency_index
                    - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    - Mobility_parks: % to baseline in visits (parks etc.)
                    - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    - Mobility_residential: % to baseline in visits (places of residence)
                    - Mobility_workplaces: % to baseline in visits (places of work)

        Note:
            When @country is None, country-level data will be returned.

        Note:
            When @country is a string and @province is None, province-level data in the country will be returned.

        Note:
            When @country and @province are strings, city-level data in the province will be returned.
        """
        db_dict = {
            "japan": _CSJapan,
            "covid19dh": _COVID19dh,
            "google": _GoogleOpenData,
            "owid": _OWID,
        }
        all_databases = list(db_dict.keys())
        selected = Validator(databases, "databases").sequence(default=all_databases, candidates=all_databases)
        for database in selected:
            db = db_dict[database](
                directory=self._directory, update_interval=self._update_interval, verbose=self._verbose)
            new_df = db.layer(country=country, province=province).convert_dtypes()
            if new_df.empty:
                continue
            self._gis.register(
                data=new_df, layers=self.LAYERS, date=self.DATE, citations=db.CITATION, convert_iso3=False)
        try:
            return self._gis.layer(geo=(country, province))
        except NotRegisteredError:
            raise SubsetNotFoundError(geo=(country, province)) from None

    def citations(self, variables=None):
        """
        Return citation list of the data sources.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)

        Returns:
            list[str]: citation list
        """
        return self._gis.citations(variables=variables)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import VisualizeBase, find_args


class BarPlot(VisualizeBase):
    """
    Create a bar plot.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._variables = []
        self._ax = None

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, vertical=True, colormap=None, color_dict=None, **kwargs):
        """
        Create bar plot.

        Args:
            data (pandas.DataFrame or pandas.Series): data to show
                Index
                    labels of the bars
                Columns
                    variables to show
            vertical (bool): whether vertical bar plot (True) or horizontal bar plot (False)
            colormap (str, matplotlib colormap object or None): colormap, please refer to https://matplotlib.org/examples/color/colormaps_reference.html
            color_dict (dict[str, str] or None): dictionary of column names (keys) and colors (values)
            kwargs: keyword arguments of pandas.DataFrame.plot()
        """
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data)
        Validator(data, "data").dataframe()
        self._variables = data.columns.tolist()
        # Color
        color_args = self._plot_colors(data.columns, colormap=colormap, color_dict=color_dict)
        # Set plotting
        method_dict = {True: data.plot.bar, False: data.plot.barh}
        try:
            self._ax = method_dict[vertical](**color_args, **kwargs)
        except ValueError as e:
            raise ValueError(e.args[0]) from None
        # No rotation of xticks
        self._ax.tick_params(axis="x", rotation=0)

    def x_axis(self, xlabel=None):
        """
        Set x axis.

        Args:
            xlabel (str or None): x-label
        """
        # Label
        self._ax.set_xlabel(xlabel)

    def y_axis(self, ylabel="Cases", y_logscale=False, ylim=(0, None), math_scale=True, y_integer=False):
        """
        Set x axis.

        Args:
            ylabel (str or None): y-label
            y_logscale (bool): whether use log-scale in y-axis or not
            ylim (tuple(int or float, int or float)): limit of y domain
            math_scale (bool): whether use LaTEX or not in y-label
            y_integer (bool): whether force to show the values as integer or not

        Note:
            If None is included in ylim, the values will be automatically determined by Matplotlib
        """
        # Label
        self._ax.set_ylabel(ylabel)
        # Math scale
        if math_scale:
            self._ax.yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
            self._ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        # Integer scale
        if y_integer:
            fmt = ScalarFormatter(useOffset=False)
            fmt.set_scientific(False)
            self._ax.yaxis.set_major_formatter(fmt)
        # Log scale
        if y_logscale:
            self._ax.set_yscale("log")
            ylim = (None, None) if ylim[0] == 0 else ylim
        # limit
        self._ax.set_ylim(*ylim)

    def line(self, v=None, h=None, color="black", linestyle=":"):
        """
        Show vertical/horizontal lines.

        Args:
            v (list[int/float] or None): list of x values of vertical lines or None
            h (list[int/float] or None): list of y values of horizontal lines or None
            color (str): color of the line
            linestyle (str): linestyle
        """
        if h is not None:
            self._ax.axhline(y=h, color=color, linestyle=linestyle)
        if v is not None:
            v = v if isinstance(v, list) else [v]
            for value in v:
                self._ax.axvline(x=value, color=color, linestyle=linestyle)


def bar_plot(df, title=None, filename=None, show_legend=True, **kwargs):
    """
    Wrapper function: show chronological change of the data.

    Args:
        data (pandas.DataFrame or pandas.Series): data to show
            Index
                Date (pandas.Timestamp)
            Columns
                variables to show
        title (str): title of the figure
        filename (str or None): filename to save the figure or None (display)
        show_legend (bool): whether show legend or not
        kwargs: keyword arguments of the following classes and methods.
            - covsirphy.BarPlot() and its methods,
            - matplotlib.pyplot.savefig(), matplotlib.pyplot.legend(),
            - pandas.DataFrame.plot()
    """
    with BarPlot(filename=filename, **find_args(plt.savefig, **kwargs)) as bp:
        bp.title = title
        bp.plot(data=df, **find_args([BarPlot.plot, pd.DataFrame.plot], **kwargs))
        # Axis
        bp.x_axis(**find_args([BarPlot.x_axis], **kwargs))
        bp.y_axis(**find_args([BarPlot.y_axis], **kwargs))
        # Vertical/horizontal lines
        bp.line(**find_args([BarPlot.line], **kwargs))
        # Legend
        if show_legend:
            bp.legend(**find_args([BarPlot.legend, plt.legend], **kwargs))
        else:
            bp.legend_hide()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from datetime import timedelta
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, NAFoundError, UnExecutedError, UnExpectedValueError
from covsirphy.util.argument import find_args
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.trend_plot import trend_plot
from covsirphy._deprecated.trend_detector import TrendDetector
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.ode_handler import ODEHandler


class PhaseTracker(Term):
    """
    Track phase information of one scenario.

    Args:
        data (pandas.DataFrame):
            Index
                reset index
            Columns
                - Date (pandas.Timestamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - Susceptible (int): the number of susceptible cases
        today (str or pandas.Timestamp): reference date to determine whether a phase is a past phase or not
        area (str): area name, like Japan/Tokyo

    Note:
        (Internally) ID=0 means not registered, ID < 0 means disabled, IDs (>0) are active phase ID.
    """

    @deprecate(old="PhaseTracker", version="2.24.0-xi")
    def __init__(self, data, today, area):
        self._ensure_dataframe(data, name="data", columns=self.SUB_COLUMNS)
        self._today = Validator(today, "today").date()
        self._area = str(area)
        # Tracker of phase information: index=Date, records of C/I/F/R/S, phase ID (0: not defined)
        self._track_df = data.set_index(self.DATE)
        self._track_df[self.ID] = 0
        # For simulation (determined in self.estimate())
        self._model = None
        self._tau = None

    def __len__(self):
        """
        int: the number of registered phases, including deactivated phases
        """
        df = self._track_df.copy()
        df = df.loc[df[self.ID] != 0]
        return df[self.ID].nunique()

    @staticmethod
    def _ensure_dataframe(target, name="df", time_index=False, columns=None, empty_ok=True):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Returns:
            pandas.DataFrame:
                Index
                    as-is
                Columns:
                    columns specified with @columns or all columns of @target (when @columns is None)
        """
        if not isinstance(target, pd.DataFrame):
            raise TypeError(f"@{name} must be a instance of (pandas.DataFrame).")
        df = target.copy()
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if not empty_ok and target.empty:
            raise ValueError(f"@{name} must not be a empty dataframe.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(f"{s1} {cols_str} must be included.")
        return df.loc[:, columns]

    @staticmethod
    def _ensure_list(target, candidates=None, name="target"):
        """
        Ensure the target is a sub-list of the candidates.

        Args:
            target (list[object]): target to ensure
            candidates (list[object] or None): list of candidates, if we have
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        if not isinstance(target, (list, tuple)):
            raise TypeError(f"@{name} must be a list or tuple, but {type(target)} was applied.")
        if candidates is None:
            return target
        # Check the target is a sub-list of candidates
        try:
            strings = [str(candidate) for candidate in candidates]
        except TypeError:
            raise TypeError(f"@candidates must be a list, but {candidates} was applied.") from None
        ok_list = [element in candidates for element in target]
        if all(ok_list):
            return target
        candidate_str = ", ".join(strings)
        raise KeyError(f"@{name} must be a sub-list of [{candidate_str}], but {target} was applied.") from None

    def define_phase(self, start, end):
        """
        Define an active phase with the series of dates.

        Args:
            start (str or pandas.Timestamp): start date of the new phase
            end (str or pandas.Timestamp): end date of the new phase

        Returns:
            covsirphy.PhaseTracker: self

        Note:
            When today is in the range of (start, end), a past phase and a future phase will be created.
        """
        start = Validator(start, name="start").date()
        end = Validator(end, name="end").date()
        track_df = self._track_df.copy()
        # Start date must be over the first date of records
        self._ensure_date_order(track_df.index.min(), start, name="start")
        # Add a past phase (start -> min(end, today))
        if start <= self._today:
            track_df.loc[start:min(self._today, end), self.ID] = track_df[self.ID].abs().max() + 1
        # Add a future phase (tomorrow -> end)
        if self._today < end:
            phase_start = max(self._today + timedelta(days=1), start)
            df = pd.DataFrame(index=pd.date_range(phase_start, end), columns=track_df.columns)
            df.index.name = self.DATE
            df[self.ID] = track_df[self.ID].abs().max() + 1
            track_df = pd.concat([track_df, df], axis=0).resample("D").last()
        # Fill in skipped dates
        series = track_df[self.ID].copy()
        track_df.loc[(series.index <= end) & (series == 0), self.ID] = series.abs().max() + 1
        # Update self
        self._track_df = track_df.copy()
        return self

    @classmethod
    def _ensure_date_order(cls, previous_date, following_date, name="following_date"):
        """
        Ensure that the order of dates.

        Args:
            previous_date (str or pandas.Timestamp): previous date
            following_date (str or pandas.Timestamp): following date
            name (str): name of @following_date

        Raises:
            ValueError: @previous_date > @following_date
        """
        previous_date = cls._ensure_date(previous_date)
        following_date = cls._ensure_date(following_date)
        p_str = previous_date.strftime(cls.DATE_FORMAT)
        f_str = following_date.strftime(cls.DATE_FORMAT)
        if previous_date <= following_date:
            return None
        raise ValueError(f"@{name} must be the same as/over {p_str}, but {f_str} was applied.")

    def deactivate(self, start, end):
        """
        The status (enabled/disabled) of the phases between start and end date will be reversed.

        Args:
            start (str or pandas.Timestamp): start date of the phase to remove
            end (str or pandas.Timestamp): end date of the phase to remove

        Returns:
            covsirphy.PhaseTracker: self
        """
        start = Validator(start, "start").date()
        end = Validator(end, "end").date()
        self._track_df.loc[start:end, self.ID] *= -1
        return self

    def remove_phase(self, start, end):
        """
        Remove phase information from the date range.

        Args:
            start (str or pandas.Timestamp): start date of the phase to remove
            end (str or pandas.Timestamp): end date of the phase to remove

        Returns:
            covsirphy.PhaseTracker: self
        """
        start = Validator(start, "start").date()
        end = Validator(end, "end").date()
        self._track_df.loc[start:end, self.ID] = 0
        return self

    def track(self):
        """
        Track data with all dates.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Susceptible (int): the number of susceptible cases
                    - If available,
                        - Rt (float): phase-dependent reproduction number
                        - (str, float): estimated parameter values, including rho
                        - (int or float): day parameters, including 1/beta [days]
                        - {metric}: score with the estimated parameter values
                        - Trials (int): the number of trials
                        - Runtime (str): runtime of optimization

        Note:
            C/I/F/R/S/I is simulated values if parameter values are available.
        """
        df = self._track_df.copy()
        df = df.loc[df[self.ID] != 0]
        # Use simulated data for tracking
        with contextlib.suppress(UnExecutedError):
            df.update(self.simulate().set_index(self.DATE))
        return df.drop(self.ID, axis=1).reset_index()

    def summary(self):
        """
        Summarize phase information.

        Returns:
            pandas.DataFrame
                Index
                    str: phase names
                Columns
                    - Type: 'Past' or 'Future'
                    - Start: start date of the phase
                    - End: end date of the phase
                    - Population: population value of the start date
                    - If available,
                        - ODE (str): ODE model names
                        - Rt (float): phase-dependent reproduction number
                        - (str, float): estimated parameter values, including rho
                        - tau (int): tau value [min]
                        - (int or float): day parameters, including 1/beta [days]
                        - {metric}: score with the estimated parameter values
                        - Trials (int): the number of trials
                        - Runtime (str): runtime of optimization
        """
        # Remove un-registered phase
        track_df = self._track_df.reset_index()
        track_df["ID_ordered"], _ = track_df[self.ID].factorize()
        track_df = track_df.loc[track_df[self.ID] > 0].drop(self.ID, axis=1)
        # -> index=phase names, columns=Start/variables,.../End
        first_df = track_df.groupby("ID_ordered").first()
        df = first_df.join(track_df.groupby("ID_ordered").last(), rsuffix="_last")
        df = df.rename(columns={self.DATE: self.START, f"{self.DATE}_last": self.END})
        df.index.name = None
        df.index = [self.num2str(num) for num in df.index]
        df = df.loc[:, [col for col in df.columns if "_last" not in col]]
        # Calculate phase types: Past or Future
        df[self.TENSE] = (df[self.START] <= self._today).map({True: self.PAST, False: self.FUTURE})
        # Calculate population values
        df[self.N] = df[[self.S, self.C]].sum(axis=1).replace(0.0, np.nan).ffill().astype(np.int64)
        # Fill in blanks of ODE model name and tau
        if self.ODE in df:
            df[self.ODE] = df[self.ODE].ffill()
        if self.TAU in df:
            df[self.TAU] = self._tau
        # Set the order of columns
        df = df.drop([self.C, self.CI, self.F, self.R, self.S], axis=1)
        fixed_cols = self.TENSE, self.START, self.END, self.N
        others = [col for col in df.columns if col not in set(fixed_cols)]
        return df.loc[:, [*fixed_cols, *others]]

    def trend(self, force, show_figure, **kwargs):
        """
        Define past phases with S-R trend analysis.

        Args:
            force (bool): if True, change points will be over-written
            show_figure (bool): if True, show the result as a figure
            kwargs: keyword arguments of covsirphy.TrendDetector(), .TrendDetector.sr() and .trend_plot()

        Returns:
            covsirphy.PhaseTracker: self
        """
        df = self._track_df.loc[:self._today].reset_index()[self.SUB_COLUMNS]
        detector = TrendDetector(data=df, area=self._area, **find_args(TrendDetector, **kwargs))
        # Perform S-R trend analysis
        detector.sr(**find_args(TrendDetector.sr, **kwargs))
        # Register phases
        if force:
            start_dates, end_dates = detector.dates()
            _ = [self.define_phase(start, end) for (start, end) in zip(start_dates, end_dates)]
        # Show S-R plane
        if show_figure:
            detector.show(**find_args(trend_plot, **kwargs))
        return self

    def estimate(self, model, tau=None, **kwargs):
        """
        Perform parameter estimation for each phases and update parameter values.

        Args:
            model (covsirphy.ModelBase): ODE model
            tau (int or None): tau value [min] or None (to be estimated)
            kwargs: keyword arguments of ODEHandler(), ODEHandler.estimate_tau() and .estimate_param()

        Returns:
            int: applied or estimated tau value [min]

        Note:
            ODE parameter estimation will be done for all active phases.
        """
        Validator(model, "model").subclass(ModelBase)
        Validator(tau, "tau").tau(default=None)
        # Set-up ODEHandler
        data_df = self._track_df.reset_index()
        data_df = data_df.loc[data_df[self.ID] > 0].dropna(how="all", axis=0)
        handler = ODEHandler(model, data_df[self.DATE].min(), tau=tau, **find_args(ODEHandler, **kwargs))
        start_dates = data_df.groupby(self.ID).first()[self.DATE].sort_values()
        end_dates = data_df.groupby(self.ID).last()[self.DATE].sort_values()
        for (start, end) in zip(start_dates, end_dates):
            y0_series = model.convert(data_df.loc[data_df[self.DATE] >= start], tau=None).iloc[0]
            _ = handler.add(end, y0_dict=y0_series.to_dict())
        # Estimate tau value if necessary
        if tau is None:
            tau = handler.estimate_tau(data_df, **find_args(ODEHandler.estimate_tau, **kwargs))
        # Estimate ODE parameter values
        est_dict = handler.estimate_params(data_df, **kwargs)
        # Register phase information to self
        df = pd.DataFrame.from_dict(est_dict, orient="index")
        df[self.DATE] = df[[self.START, self.END]].apply(lambda x: pd.date_range(x[0], x[1]), axis=1)
        df = df.explode(self.DATE).drop([self.START, self.END], axis=1).set_index(self.DATE)
        df.insert(0, self.ODE, model.NAME)
        df.insert(6, self.TAU, tau)
        all_columns = [*self._track_df.columns.tolist(), *df.columns.tolist()]
        sorted_columns = sorted(set(all_columns), key=all_columns.index)
        self._track_df = self._track_df.combine_first(df)
        self._track_df = self._track_df.reindex(columns=sorted_columns)
        # Set model and tau to self
        self._model = model
        self._tau = tau
        return tau

    def set_ode(self, model, param_df, tau):
        """
        Set ODE model, parameter values manually, not using parameter estimation.

        Args:
            model (covsirphy.ModelBase): ODE model
            param_df (pandas.DataFrame):
                Index
                    Date (pandas.Timestamp): dates to update parameter values
                Columns
                    (float): parameter values
            tau (int): tau value [min] (must not be None)

        Raises:
            ValueError: some model parameters are not included in @param_df

        Note:
            Parameters are defined by model.PARAMETERS.

        Returns:
            int: applied tau value [min]

        Note:
            ODE model for simulation will be overwritten.
        """
        self._model = Validator(model, "model").subclass(ModelBase)
        self._ensure_dataframe(param_df, name="param_df", time_index=True, columns=model.PARAMETERS)
        self._tau = Validator(tau, "tau").tau(default=None)
        if self._tau is None:
            raise NAFoundError("tau")
        new_df = param_df.copy()
        # Add model name
        new_df.insert(0, self.ODE, model.NAME)
        # Calculate reproduction number
        new_df.insert(1, self.RT, None)
        new_df[self.RT] = new_df[model.PARAMETERS].apply(
            lambda x: model(1, **x.to_dict()).calc_r0(), axis=1)
        # Add tau
        new_df[self.TAU] = self._tau
        # Calculate days parameters
        days_df = new_df[model.PARAMETERS].apply(
            lambda x: pd.Series(model(1, **x.to_dict()).calc_days_dict(self._tau)), axis=1)
        new_df = pd.concat([new_df, days_df], axis=1)
        # update tracker
        columns_include_dup = [*self._track_df.columns.tolist(), *new_df.columns.tolist()]
        track_df = self._track_df.reindex(
            columns=sorted(set(columns_include_dup), key=columns_include_dup.index))
        track_df.update(new_df)
        self._track_df = track_df.copy()
        return self._tau

    def simulate(self):
        """
        Perform simulation with the multi-phased ODE model.

        Raises:
            covsirphy.UnExecutedError: either tau value or phase information was not set

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Susceptible (int): the number of susceptible cases

        Note:
            Deactivated phases will be included.

        Note:
            Un-registered phases will not be included.

        Note:
            If parameter set is not registered for the current phase and
            the previous phase has parameter set, this set will be used for the current phase.
        """
        # Model and tau must be set
        if self._model is None:
            raise UnExecutedError("PhaseTracker.estimate() or PhaseTracker.set_ode()")
        # Get parameter sets and initial values
        record_df = self._track_df.copy()
        record_df = record_df.loc[record_df[self.ID] != 0].ffill().dropna(subset=self._model.PARAMETERS)
        start_dates = record_df.reset_index().groupby(self.ID).first()[self.DATE].sort_values()
        end_dates = record_df.reset_index().groupby(self.ID).last()[self.DATE].sort_values()
        # Set-up ODEHandler
        handler = ODEHandler(self._model, record_df.index.min(), tau=self._tau)
        parameters = self._model.PARAMETERS[:]
        for (start, end) in zip(start_dates, end_dates):
            param_dict = record_df.loc[end, parameters].to_dict()
            if end <= self._today:
                ph_df = record_df.loc[start:, [self.S, self.CI, self.F, self.R]].reset_index()
                y0_dict = self._model.convert(ph_df, self._tau).iloc[0].to_dict()
            else:
                y0_dict = None
            _ = handler.add(end, param_dict=param_dict, y0_dict=y0_dict)
        # Perform simulation
        sim_df = handler.simulate()
        sim_df[self.C] = sim_df[[self.CI, self.F, self.R]].sum(axis=1)
        return sim_df.loc[:, self.SUB_COLUMNS]

    def parse_range(self, dates=None, past_days=None, phases=None):
        """
        Parse date range and return the minimum date and maximum date.

        Args:
            dates (tuple(str or pandas.Timestamp or None, ) or None): start date and end date
            past_days (int or None): how many past days to use in calculation from today (property)
            phases (list[str] or None): phase names to use in calculation

        Raises:
            covsirphy.UnExecutedError: no phases were registered
            ValueError: @dates argument does not have exact two elements

        Returns:
            tuple(pandas.Timestamp, pandas.Timestamp): the minimum date and maximum date

        Notes:
            When not specified (i.e. None was applied),
            the start date of the 0th phase will be used as the minimum date.

        Notes:
            When not specified (i.e. None was applied),
            the end date of the last phase phase will be used as the maximum date.

        Note:
            When @past_days was specified, (today - @past_days, today) will be returned.

        Note:
            In @phases, 'last' means the last registered phase.

        Note:
            Priority is given in the order of @dates, @past_days, @phases.
        """
        if not self:
            raise UnExecutedError("PhaseTracker.define_phase()")
        # Get list of phases: index=phase names, columns=Start/End
        track_df = self._track_df.reset_index()
        track_df = track_df.loc[track_df[self.ID] != 0]
        track_df[self.ID], _ = track_df[self.ID].factorize()
        first_df = track_df.groupby(self.ID).first()
        df = first_df.join(track_df.groupby(self.ID).last(), rsuffix="_last")
        df = df.rename(columns={self.DATE: self.START, f"{self.DATE}_last": self.END})
        df.index = [self.num2str(num) for num in df.index]
        # Get default values
        start_default, end_default = df[self.START].min(), df[self.END].max()
        # Read @dates
        if dates is not None:
            if len(dates) != 2:
                raise ValueError(f"@dates must be a tuple which has two elements, but {dates} was applied.")
            start = Validator(dates[0], name="the first element of 'dates' argument").date(default=start_default)
            end = Validator(dates[1], name="the second element of 'dates' argument").date(default=end_default)
            self._ensure_date_order(start, end, name="the second element of 'dates' argument")
            return (start, end)
        # Read @past_days
        if past_days is not None:
            past_days = Validator(past_days, "past_days").int(value_range=(1, None))
            return (self._today - timedelta(days=past_days), self._today)
        # No arguments were specified
        if phases is None:
            return (start_default, end_default)
        # Read @phases
        self._ensure_list(phases, name="phases")
        dates = []
        for phase in phases:
            phase_replaced = df.index[-1] if phase == "last" else phase
            self._ensure_selectable(phase_replaced, df.index.tolist(), name="phase")
            start = df.loc[phase_replaced, self.START]
            end = df.loc[phase_replaced, self.END]
            dates.extend(pd.date_range(start, end).tolist())
        return (min(dates), max(dates))

    def _ensure_selectable(self, target, candidates, name="target"):
        """
        Ensure that the target can be selectable.

        Args:
            target (object): target to check
            candidates (list[object]): list of candidates
            name (str): name of the target
        """
        Validator(candidates, "candidates").sequence()
        if target in candidates:
            return target
        raise UnExpectedValueError(name=name, value=target, candidates=candidates)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import VisualizeBase, find_args


class ComparePlot(VisualizeBase):
    """
    Compare two groups with specified variables.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._variables = []
        self._ax = None

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, variables, groups):
        """
        Compare two groups with specified variables.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    x values
                Columns
                    y variables to show, "{variable}_{group}" for all combinations of variables and groups
            variables (list[str]): variables to compare
            groups (list[str]): the first group name and the second group name
        """
        Validator(variables, "variables").sequence()
        group1, group2 = Validator(groups, "groups").sequence()
        col_nest = [[f"{variable}_{group}" for group in groups] for variable in variables]
        Validator(data, "data").dataframe(columns=sum(col_nest, []))
        # Prepare figure object
        fig_len = len(variables) + 1
        _, self._ax = plt.subplots(ncols=1, nrows=fig_len, figsize=(9, 6 * fig_len / 2))
        # Compare each variable
        for (ax, v, columns) in zip(self._ax.ravel()[1:], variables, col_nest):
            data[columns].plot.line(
                ax=ax, ylim=(None, None), sharex=True, title=f"Comparison regarding {v}(t)")
            ax.yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
            ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
            ax.legend(bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0)
        # Show residuals
        for (v, columns) in zip(variables, col_nest):
            data[f"{v}_diff"] = data[columns[0]] - data[columns[1]]
            data[f"{v}_diff"].plot.line(
                ax=self._ax.ravel()[0], sharex=True,
                title=f"{group1.capitalize()} - {group2.capitalize()}")
        self._ax.ravel()[0].axhline(y=0, color="black", linestyle="--")
        self._ax.ravel()[0].yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
        self._ax.ravel()[0].ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        self._ax.ravel()[0].legend(bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0)


def compare_plot(df, variables, groups, filename=None, **kwargs):
    """
    Wrapper function: show chronological change of the data.

    Args:
        df (pandas.DataFrame): data to show
            Index
                x values
            Columns
                y variables to show, "{variable}_{group}" for all combinations of variables and groups
        variables (list[str]): variables to compare
        groups (list[str]): the first group name and the second group name
        filename (str or None): filename to save the figure or None (display)
        kwargs: keyword arguments of the following classes and methods.
            - matplotlib.pyplot.savefig()
            - matplotlib.pyplot.legend()
    """
    with ComparePlot(filename=filename, **find_args(plt.savefig, **kwargs)) as cp:
        cp.plot(data=df, variables=variables, groups=groups)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.error import deprecate
from covsirphy.util.term import Term


class PhaseSeries(Term):
    """
    A series of phases.

    Args:
        first_date (str): the first date of the series, like 22Jan2020
        last_date (str): the last date of the records, like 25May2020
        population (int): initial value of total population in the place
    """

    @deprecate("PhaseSeries", new="ODEHandler", version="2.19.1-zeta-fu1")
    def __init__(self, first_date, last_date, population):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.term import Term
from covsirphy.downloading.db import _DataBase


class _COVID19dh(_DataBase):
    """
    Access "COVID-19 Data Hub" server.
    https://covid19datahub.io/
    """
    # File title without extensions and suffix
    TITLE = "covid19dh"
    # Dictionary of column names
    _OXCGRT_COLS_RAW = [
        "school_closing",
        "workplace_closing",
        "cancel_events",
        "gatherings_restrictions",
        "transport_closing",
        "stay_home_restrictions",
        "internal_movement_restrictions",
        "international_movement_restrictions",
        "information_campaigns",
        "testing_policy",
        "contact_tracing",
        "stringency_index",
    ]
    OXCGRT_VARS = [v.capitalize() for v in _OXCGRT_COLS_RAW]
    COL_DICT = {
        "date": Term.DATE,
        "iso_alpha_3": Term.ISO3,
        "administrative_area_level_1": Term.COUNTRY,
        "administrative_area_level_2": Term.PROVINCE,
        "administrative_area_level_3": Term.CITY,
        "tests": Term.TESTS,
        "confirmed": Term.C,
        "deaths": Term.F,
        "recovered": Term.R,
        "population": Term.N,
        **dict(zip(_OXCGRT_COLS_RAW, OXCGRT_VARS)),
    }
    # Stdout when downloading (shown at most one time)
    STDOUT = "Retrieving datasets from COVID-19 Data Hub https://covid19datahub.io/"
    # Citation
    CITATION = 'Guidotti, E., Ardia, D., (2020), "COVID-19 Data Hub",' \
        ' Journal of Open Source Software 5(51):2376, doi: 10.21105/joss.02376.'

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (pandas.NA): NAs
                    - City (pandas.NA): NAs
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Population (numpy.int64): population values
                    - Tests (numpy.float64): the number of tests
        """
        url = "https://storage.covid19datahub.io/level/1.csv.zip"
        df = self._provide(
            url=url, suffix="_level1", columns=list(self.COL_DICT.keys()), date="date", date_format="%Y-%m-%d")
        # ships will be regarded as provinces of "Others" country
        ships = df.loc[df[self.ISO3].isna(), self.COUNTRY].unique()
        for ship in ships:
            df.loc[df[self.COUNTRY] == ship, [self.ISO3, self.PROVINCE]] = [self.OTHERS, ship]
        return df

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (pandas.NA): NAs
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Population (numpy.int64): population values
                    - Tests (numpy.float64): the number of tests
        """
        iso3 = self._to_iso3(country)[0]
        if iso3 == self.OTHERS:
            df = self._country()
            return df.loc[df[self.ISO3] == iso3]
        url = f"https://storage.covid19datahub.io/country/{iso3}.csv.zip"
        df = self._provide(
            url=url, suffix=f"_{iso3.lower()}", columns=list(self.COL_DICT.keys()), date="date", date_format="%Y-%m-%d")
        df = df.loc[(~df[self.PROVINCE].isna()) & (df[self.CITY].isna())]
        df.loc[:, self.CITY] = df.loc[:, self.CITY].fillna(self.NA)
        return df

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Population (numpy.int64): population values
                    - Tests (numpy.float64): the number of tests
        """
        iso3 = self._to_iso3(country)[0]
        url = f"https://storage.covid19datahub.io/country/{iso3}.csv.zip"
        df = self._provide(
            url=url, suffix=f"_{iso3.lower()}", columns=list(self.COL_DICT.keys()), date="date", date_format="%Y-%m-%d")
        return df.loc[(df[self.PROVINCE] == province) & (~df[self.CITY].isna())]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from copy import deepcopy
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _SubsetManager(Term):
    """
    Class to get subset with location information.

    Args:
        layers (list[str]): names of administration layers with the order (upper layers precede, e.g. ["Country", "Province"])
    """

    def __init__(self, layers):
        self._layers = Validator(layers, "layers").sequence(candidates=None)

    def layer(self, data, geo=None):
        """Return the data at the selected layer.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - columns defined by `SubsetManager(layers)` argument: note that "-" means total values of the upper layer
                    - the other columns of values
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to filter or None (top-level layer)

        Raises:
            TypeError: @geo has un-expected types
            ValueError: the length of @geo is equal to or larger than the length of layers

        Returns:
            pandas.DataFrame: as-is @data

        Note:
           Note that records with NAs as country names will be always removed.

        Note:
            When `geo=None` or `geo=(None,)`, returns country-level data, assuming we have country/province/city as layers here.

        Note:
            When `geo=("Japan",)` or `geo="Japan"`, returns province-level data in Japan.

        Note:
            When `geo=(["Japan", "UK"],)`, returns province-level data in Japan and UK.

        Note:
            When `geo=("Japan", "Kanagawa")`, returns city-level data in Kanagawa/Japan.

        Note:
            When `geo=("Japan", ["Tokyo", "Kanagawa"])`, returns city-level data in Tokyo/Japan and Kanagawa/Japan.
        """
        if geo is not None and not isinstance(geo, (list, tuple, str)):
            raise TypeError(
                f"@geo must be a tuple(list[str] or tuple(str) or str) or str or None, but {geo} was applied.")
        geo_converted = (geo,) if (geo is None or isinstance(geo, str)) else deepcopy(geo)
        df = Validator(data, "data").dataframe(columns=self._layers, empty_ok=False)
        df.loc[:, self._layers] = df[self._layers].fillna(self.NA)
        df = df.loc[df[self._layers[0]] != self.NA]
        for (i, sel) in enumerate(geo_converted):
            if sel is None:
                try:
                    return df.loc[df[self._layers[i + 1]] == self.NA].reset_index(drop=True)
                except IndexError:
                    return df.reset_index(drop=True)
            if not isinstance(sel, (str, list, tuple)):
                raise TypeError(f"@geo must be a tuple(list[str] or tuple(str) or str) or None, but {geo} was applied.")
            if i >= len(self._layers):
                raise ValueError(f"The length of @geo must be smaller than that of layers, but {geo} was applied.")
            df = df.loc[df[self._layers[i]].isin([sel] if isinstance(sel, str) else sel)]
            if i == len(geo_converted) - 1 and i < len(self._layers) - 1:
                df = df.loc[df[self._layers[i + 1]] != self.NA]
                with contextlib.suppress(IndexError):
                    df = df.loc[df[self._layers[i + 2]] == self.NA]
        return df.reset_index(drop=True)

    def filter(self, data, geo=None):
        """Filter the data with geography information.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - columns defined by `SubsetManager(layers)` argument: note that "-" means total values of the upper layer
                    - the other columns of values
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names for the layers to filter or None (all data at the top level)

        Raises:
            TypeError: @geo has un-expected types
            ValueError: the length of @geo is larger than the length of layers

        Returns:
            pandas.DataFrame: as-is @data

        Note:
           Note that records with NAs as country names will be always removed.

        Note:
            When `geo=None` or `geo=(None,)`, returns all country-level data, assuming we have country/province/city as layers here.

        Note:
            When `geo=("Japan",)` or `geo="Japan"`, returns country-level data in Japan.

        Note:
            When `geo=(["Japan", "UK"],)`, returns country-level data of Japan and UK.

        Note:
            When `geo=("Japan", "Tokyo")`, returns province-level data of Tokyo/Japan.

        Note:
            When `geo=("Japan", ["Tokyo", "Kanagawa"])`, returns province-level data of Tokyo/Japan and Kanagawa/Japan.

        Note:
            When `geo=("Japan", "Kanagawa", "Yokohama")`, returns city-level data of Yokohama/Kanagawa/Japan.

        Note:
            When `geo=(("Japan", "Kanagawa", ["Yokohama", "Kawasaki"])`, returns city-level data of Yokohama/Kanagawa/Japan and Kawasaki/Kanagawa/Japan.
        """
        if geo is None or geo == (None,) or geo == [None]:
            return self.layer(data=data, geo=None)
        if not isinstance(geo, (list, tuple, str)):
            raise TypeError(
                f"@geo must be a tuple(list[str] or tuple(str) or str) or str or None, but {geo} was applied.")
        geo_converted = (geo,) if isinstance(geo, str) else deepcopy(geo)
        if len(geo_converted) > len(self._layers):
            raise ValueError(f"The length of @geo cannot be larger than that of layers, but {geo} was applied.")
        *geo_formers, geo_last = geo_converted
        df = self.layer(data=data, geo=geo_formers or None)
        if not isinstance(geo_last, (str, list, tuple)):
            raise TypeError(
                f"The last value of @geo must be a list[str] or tuple(str) or str, but {geo_last} was applied.")
        selectors = [geo_last] if isinstance(geo_last, str) else geo_last
        return df.loc[df[self._layers[len(geo_converted) - 1]].isin(selectors)].reset_index(drop=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.validator import Validator
from covsirphy.dynamics.ode import ODEModel


class SIRModel(ODEModel):
    """Class of SIR model.

    Args:
        date_range (tuple of (str, str)): start date and end date of simulation
        tau (int): tau value [min]
        initial_dict (dict of {str: int}): initial values
            - Susceptible (int): the number of susceptible cases
            - Infected (int): the number of infected cases
            - Fatal or Recovered (int): the number of fatal or recovered cases
        param_dict (dict of {str: float}): non-dimensional parameter values
            - rho: non-dimensional effective contact rate
            - sigma: non-dimensional recovery plus mortality rate
    """
    # Name of ODE model
    _NAME = "SIR Model"
    # Variables
    _VARIABLES = [ODEModel.S, ODEModel.CI, ODEModel.FR]
    # Non-dimensional parameters
    _PARAMETERS = ["rho", "sigma"]
    # Dimensional parameters
    _DAY_PARAMETERS = ["1/beta [day]", "1/gamma [day]"]
    # Variables that increases monotonically
    _VARS_INCREASE = [ODEModel.FR]
    # Sample data
    _SAMPLE_DICT = {
        "initial_dict": {ODEModel.S: 999_000, ODEModel.CI: 1000, ODEModel.FR: 0},
        "param_dict": {"rho": 0.2, "sigma": 0.075}
    }

    def __init__(self, date_range, tau, initial_dict, param_dict):
        super().__init__(date_range, tau, initial_dict, param_dict)
        self._rho = Validator(self._param_dict["rho"], "rho", accept_none=False).float(value_range=(0, 1))
        self._sigma = Validator(self._param_dict["sigma"], "sigma", accept_none=False).float(value_range=(0, 1))

    def _discretize(self, t, X):
        """Discretize the ODE.

        Args:
            t (int): discrete time-steps
            X (numpy.array): the current values of the model

        Returns:
            numpy.array: the next values of the model
        """
        n = self._population
        s, i, *_ = X
        dsdt = 0 - self._rho * s * i / n
        drdt = self._sigma * i
        didt = 0 - dsdt - drdt
        return np.array([dsdt, didt, drdt])

    @classmethod
    def transform(cls, data, tau=None):
        """Transform a dataframe, converting Susceptible/Infected/Fatal/Recovered to model-specific variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index or pandas.DatetimeIndex (when tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
            tau (int or None): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    as the same as index if @data when @tau is None else converted to time(x) = (TIME(x) - TIME(0)) / tau
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of infected cases
                    - Fatal or Recovered (int): the number of fatal or recovered cases
        """
        df = Validator(data, "data").dataframe(columns=cls._SIRF)
        df.index = cls._date_to_non_dim(df.index, tau=tau)
        df[cls.FR] = df[cls.F] + df[cls.R]
        return df.loc[:, cls._VARIABLES].convert_dtypes()

    @classmethod
    def inverse_transform(cls, data, tau=None, start_date=None):
        """Transform a dataframe, converting model-specific variables to Susceptible/Infected/Fatal/Recovered.

        Args:
            data (pandas.DataFrame):
                Index
                    any index
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of infected cases
                    - Fatal or Recovered (int): the number of fatal or recovered cases
            tau (int or None): tau value [min]
            start_date (str or pandas.Timestamp or None): start date of records ie. TIME(0)

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.DatetimeIndex) or as-is @data (when either @tau or @start_date are None the index @data is date)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
        """
        df = Validator(data, "data").dataframe(columns=cls._VARIABLES)
        df = cls._non_dim_to_date(data=df, tau=tau, start_date=start_date)
        df[cls.F] = 0
        df[cls.R] = df.loc[:, cls.FR]
        return df.loc[:, cls._SIRF].convert_dtypes()

    def r0(self):
        """Calculate basic reproduction number.

        Raises:
            ZeroDivisionError: Sigma value was over 0

        Returns:
            float: reproduction number of the ODE model and parameters
        """
        try:
            return round(self._rho / self._sigma, 2)
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Sigma must be over 0 to calculate reproduction number with {self._NAME}.") from None

    def dimensional_parameters(self):
        """Calculate dimensional parameter values.

        Raises:
            ZeroDivisionError: either rho or sigma value was over 0

        Returns:
            dict of {str: int}: dictionary of dimensional parameter values
                - "1/beta [day]" (int): infection period
                - "1/gamma [day]" (int): recovery period
        """
        try:
            return {
                "1/beta [day]": round(self._tau / 24 / 60 / self._rho),
                "1/gamma [day]": round(self._tau / 24 / 60 / self._sigma)
            }
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Rho and sigma must be over 0 to calculate dimensional parameters with {self._NAME}.") from None

    @classmethod
    def _param_quantile(cls, data, q=0.5):
        """With combinations (X, dX/dt) for X=S, I, R, calculate quantile values of ODE parameters.

        Args:
            data (pandas.DataFrame): transformed data with covsirphy.SIRModel.transform(data=data, tau=tau)
            q (float or array-like): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict of {str: float or pandas.Series}: parameter values at the quantile(s)

        Note:
            We can get approximate parameter values with difference equations as follows.
            - rho = - n * (dS/dt) / S / I
            - sigma = (dR/dt) / I
        """
        df = data.copy()
        periods = round((df.index.max() - df.index.min()) / len(df))
        # Remove negative values and set variables
        df = df.loc[(df[cls.S] > 0) & (df[cls.CI] > 0)]
        n = df.loc[df.index[0], cls._VARIABLES].sum()
        # Calculate parameter values with non-dimensional difference equation
        rho_series = 0 - n * df[cls.S].diff() / periods / df[cls.S] / df[cls.CI]
        sigma_series = df[cls.FR].diff() / periods / df[cls.CI]
        # Guess representative values
        return {
            "rho": cls._clip(rho_series.quantile(q=q), 0, 1),
            "sigma": cls._clip(sigma_series.quantile(q=q), 0, 1),
        }

    @classmethod
    def sr(cls, data):
        """Return log10(S) and R of model-specific variables for S-R trend analysis.

        Args:
            data (pandas.DataFrame):
                Index
                    Date (pd.Timestamp): Observation date
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): date
                Columns
                    log10(S) (np.float64): common logarithm of Susceptible
                    R (np.int64): Fatal or Recovered
        """
        Validator(data, "data", accept_none=False).dataframe(time_index=True, columns=cls._SIRF)
        df = data.copy()
        df[cls._logS] = np.log10(df[cls.S])
        df[cls._r] = df[cls.F] + df[cls.R]
        return df.loc[:, [cls._logS, cls._r]].astype({cls._logS: np.float64, cls._r: np.int64})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
import pandas as pd
from covsirphy.util.error import UnExecutedError
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import find_args
from covsirphy.visualization.line_plot import LinePlot


class ScatterPlot(LinePlot):
    """
    Create a scatter plot.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._ax = None
        self._data = pd.DataFrame(columns=["x", "y"])

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, colormap=None, color_dict=None, **kwargs):
        """
        Plot chronological change of the data.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    reset index
                Columns
                    x (int or float): x values
                    y (int or float): y values
            colormap (str, matplotlib colormap object or None): colormap, please refer to https://matplotlib.org/examples/color/colormaps_reference.html
            color_dict (dict[str, str] or None): dictionary of column names (keys) and colors (values)
            kwargs: keyword arguments of pandas.DataFrame.plot()
        """
        self._data = Validator(data, "data").dataframe(columns=["x", "y"])
        # Color
        color_args = self._plot_colors(data.columns, colormap=colormap, color_dict=color_dict)
        # Set plotting
        try:
            self._ax = data.plot.scatter(x="x", y="y", **color_args, **kwargs)
        except ValueError as e:
            raise ValueError(e.args[0]) from None

    def line_straight(self, p1=None, p2=None, color="black", linestyle=":"):
        """
        Connect the points with a straight line.

        Args:
            p1 (tuple(int or float, int or float) or None): (x, y) of the first point or None (min values)
            p2 (tuple(int or float, int or float) or None): (x, y) of the second point or None (max values)
            color (str): color of the line
            linestyle (str): linestyle

        Note:
            The same line will be show when p1 and p2 is reordered.
        """
        if self._data.empty:
            raise UnExecutedError("ScatterPlot.plot()")
        x1, y1 = (self._data["x"].min(), self._data["y"].min()) if p1 is None else p1
        x2, y2 = (self._data["x"].max(), self._data["y"].max()) if p2 is None else p2
        self._ax.plot([x1, x2], [y1, y2], color=color, linestyle=linestyle)

    def legend(self, **kwargs):
        """
        ScatterPlot.legend() is not implemented.
        """
        raise NotImplementedError

    def legend_hide(self):
        """
        ScatterPlot.legend_hide() is not implemented.
        """
        raise NotImplementedError


def scatter_plot(df, title=None, filename=None, **kwargs):
    """
    Wrapper function: show chronological change of the data.

    Args:
        data (pandas.DataFrame): data to show
            Index
                reset index
            Columns
                x (int or float): x values
                y (int or float): y values
        title (str): title of the figure
        filename (str or None): filename to save the figure or None (display)
        kwargs: keyword arguments of the following classes and methods.
            - covsirphy.ScatterPlot() and its methods,
            - matplotlib.pyplot.savefig(), matplotlib.pyplot.legend(),
            - pandas.DataFrame.plot()
    """
    with ScatterPlot(filename=filename, **find_args(plt.savefig, **kwargs)) as sp:
        sp.title = title
        sp.plot(data=df, **find_args([ScatterPlot.plot, pd.DataFrame.plot], **kwargs))
        # Axis
        sp.x_axis(**find_args([ScatterPlot.x_axis], **kwargs))
        sp.y_axis(**find_args([ScatterPlot.y_axis], **kwargs))
        # Vertical/horizontal lines
        sp.line(**find_args([ScatterPlot.line], **kwargs))
        # Straight lines
        sp.line_straight(**find_args([ScatterPlot.line_straight], **kwargs))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.validator import Validator
from covsirphy.dynamics.ode import ODEModel


class SEWIRFModel(ODEModel):
    """Class of SEWIR-F model.

    Args:
        date_range (tuple of (str, str)): start date and end date of simulation
        tau (int): tau value [min]
        initial_dict (dict of {str: int}): initial values
            - Susceptible (int): the number of susceptible cases
            - Exposed (int): the number of cases who are exposed and in latent period without infectivity
            - Waiting (int): the number of cases who are waiting for confirmation diagnosis with infectivity
            - Infected (int): the number of infected cases
            - Fatal (int): the number of fatal cases
            - Recovered (int): the number of recovered cases
        param_dict (dict of {str: float}): non-dimensional parameter values
            - theta: direct fatality probability of un-categorized confirmed cases
            - kappa: non-dimensional mortality rate
            - rho1: non-dimensional exposure rate (the number of encounter with the virus in a minute)
            - rho2: non-dimensional inverse value of latent period
            - rho3: non-dimensional inverse value of waiting time for confirmation
            - sigma: non-dimensional recovery rate
    """
    # Name of ODE model
    _NAME = "SEWIR-F Model"
    # Variables
    _VARIABLES = [ODEModel.S, ODEModel.E, ODEModel.W, ODEModel.CI, ODEModel.R, ODEModel.F]
    # Non-dimensional parameters
    _PARAMETERS = ["theta", "kappa", "rho1", "rho2", "rho3", "sigma"]
    # Dimensional parameters
    _DAY_PARAMETERS = [
        "alpha1 [-]", "1/alpha2 [day]", "1/beta1 [day]", "1/beta2 [day]", "1/beta3 [day]", "1/gamma [day]"]
    # Variables that increases monotonically
    _VARS_INCREASE = [ODEModel.R, ODEModel.F]
    # Sample data
    _SAMPLE_DICT = {
        "initial_dict": {
            ODEModel.S: 994_000, ODEModel.E: 3_000, ODEModel.W: 2_000, ODEModel.CI: 1000, ODEModel.R: 0, ODEModel.F: 0},
        "param_dict": {
            "theta": 0.002, "kappa": 0.005, "rho1": 0.2, "sigma": 0.075, "rho2": 0.167, "rho3": 0.167}
    }

    def __init__(self, date_range, tau, initial_dict, param_dict):
        super().__init__(date_range, tau, initial_dict, param_dict)
        self._theta = Validator(self._param_dict["theta"], "theta", accept_none=False).float(value_range=(0, 1))
        self._kappa = Validator(self._param_dict["kappa"], "kappa", accept_none=False).float(value_range=(0, 1))
        self._rho1 = Validator(self._param_dict["rho1"], "rho1", accept_none=False).float(value_range=(0, 1))
        self._rho2 = Validator(self._param_dict["rho2"], "rho2", accept_none=False).float(value_range=(0, 1))
        self._rho3 = Validator(self._param_dict["rho3"], "rho3", accept_none=False).float(value_range=(0, 1))
        self._sigma = Validator(self._param_dict["sigma"], "sigma", accept_none=False).float(value_range=(0, 1))

    def _discretize(self, t, X):
        """Discretize the ODE.

        Args:
            t (int): discrete time-steps
            X (numpy.array): the current values of the model

        Returns:
            numpy.array: the next values of the model
        """
        n = self._population
        s, i, *_, e, w = X
        beta_swi = self._rho1 * s * (w + i) / n
        dsdt = 0 - beta_swi
        dedt = beta_swi - self._rho2 * e
        dwdt = self._rho2 * e - self._rho3 * w
        drdt = self._sigma * i
        dfdt = self._kappa * i + self._theta * self._rho3 * w
        didt = 0 - dsdt - drdt - dfdt - dedt - dwdt
        return np.array([dsdt, didt, drdt, dfdt, dedt, dwdt])

    @ classmethod
    def transform(cls, data, tau=None):
        """Transform a dataframe, converting Susceptible/Infected/Fatal/Recovered to model-specific variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index or pandas.DatetimeIndex (when tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int or None): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    as the same as index if @data when @tau is None else converted to time(x) = (TIME(x) - TIME(0)) / tau
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Exposed (int): the number of cases who are exposed and in latent period without infectivity
                    - Waiting (int): the number of cases who are waiting for confirmation diagnosis with infectivity
                    - Infected (int): the number of infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
        """
        df = Validator(data, "data").dataframe(columns=cls._SIRF)
        df.index = cls._date_to_non_dim(df.index, tau=tau)
        df[cls.E] = 0
        df[cls.W] = 0
        return df.loc[:, cls._VARIABLES].convert_dtypes()

    @ classmethod
    def inverse_transform(cls, data, tau=None, start_date=None):
        """Transform a dataframe, converting model-specific variables to Susceptible/Infected/Fatal/Recovered.

        Args:
            data (pandas.DataFrame):
                Index
                    any index
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Exposed (int): the number of cases who are exposed and in latent period without infectivity
                    - Waiting (int): the number of cases who are waiting for confirmation diagnosis with infectivity
                    - Infected (int): the number of infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
            tau (int or None): tau value [min]
            start_date (str or pandas.Timestamp or None): start date of records ie. TIME(0)

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.DatetimeIndex) or as-is @data (when either @tau or @start_date are None the index @data is date)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
        """
        df = Validator(data, "data").dataframe(columns=cls._VARIABLES)
        df = cls._non_dim_to_date(data=df, tau=tau, start_date=start_date)
        df[cls.S] = df[cls.S] + df[cls.E] + df[cls.W]
        return df.loc[:, cls._SIRF].convert_dtypes()

    def r0(self):
        """Calculate basic reproduction number.

        Raises:
            ZeroDivisionError: rho2 or sigma + kappa value was over 0

        Returns:
            float: reproduction number of the ODE model and parameters
        """
        try:
            rho = self._rho1 / self._rho2 * self._rho3
            rt = rho * (1 - self._theta) / (self._sigma + self._kappa)
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Both of 'rho2' and 'sigma + kappa' must be over 0 to calculate reproduction number with {self._NAME}.") from None
        return round(rt, 2)

    def dimensional_parameters(self):
        """Calculate dimensional parameter values.

        Raises:
            ZeroDivisionError: either kappa or rho_i for i=1,2,3 or sigma value was over 0

        Returns:
            dict of {str: int}: dictionary of dimensional parameter values
                - "alpha1 [-]" (float): direct fatality probability of un-categorized confirmed cases
                - "1/alpha2 [day]" (int): mortality period of infected cases
                - "1/beta1 [day]" (int): period for susceptible people to encounter with the virus
                - "1/beta2 [day]" (int): latent period
                - "1/beta3 [day]" (int): waiting time for confirmation
                - "1/gamma [day]" (int): recovery period
        """
        try:
            return {
                "alpha1 [-]": round(self._theta, 3),
                "1/alpha2 [day]": round(self._tau / 24 / 60 / self._kappa),
                "1/beta1 [day]": round(self._tau / 24 / 60 / self._rho1),
                "1/beta2 [day]": round(self._tau / 24 / 60 / self._rho2),
                "1/beta3 [day]": round(self._tau / 24 / 60 / self._rho3),
                "1/gamma [day]": round(self._tau / 24 / 60 / self._sigma)
            }
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Kappa, rho_i for i=1,2,3 and sigma must be over 0 to calculate dimensional parameters with {self._NAME}.") from None

    @classmethod
    def from_data_with_quantile(cls, *args, **kwargs):
        """Initialize model with data, estimating ODE parameters with quantiles.

        Raises:
            NotImplementedError: this model cannot be used for parameter estimation because Exposed/Waiting data is un-available
        """
        raise NotImplementedError(
            "SEWIR-F model cannot be used for parameter estimation because we do not have records "
            "of Exposed and Waiting. Please use SIR-F model with `covsirphy.SIRFModel` class."
        )

    @classmethod
    def from_data_with_optimization(cls, *args, **kwargs):
        """Initialize model with data, estimating ODE parameters hyperparameter optimization using Optuna.

        Raises:
            NotImplementedError: this model cannot be used for parameter estimation because Exposed/Waiting data is un-available
        """
        raise NotImplementedError(
            "SEWIR-F model cannot be used for parameter estimation because we do not have records "
            "of Exposed and Waiting. Please use SIR-F model with `covsirphy.SIRFModel` class."
        )

    @classmethod
    def sr(cls, data):
        """Return log10(S) and R of model-specific variables for S-R trend analysis.

        Args:
            data (pandas.DataFrame):
                Index
                    - Date (pd.Timestamp): Observation date
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): date
                Columns
                    log10(S) (np.float64): common logarithm of Susceptible
                    R (np.int64): Recovered
        """
        Validator(data, "data", accept_none=False).dataframe(time_index=True, columns=cls._SIRF)
        df = data.rename(columns={cls.R: cls._r})
        df[cls._logS] = np.log10(df[cls.S])
        return df.loc[:, [cls._logS, cls._r]].astype({cls._logS: np.float64, cls._r: np.int64})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from itertools import groupby
from operator import itemgetter
import pandas as pd
from covsirphy.util.error import deprecate, UnExecutedError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated.population import PopulationData
from covsirphy._deprecated.oxcgrt import OxCGRTData
from covsirphy.visualization.line_plot import line_plot
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.phase_estimator import MPEstimator
from covsirphy._deprecated.scenario import Scenario


class PolicyMeasures(Term):
    """
    Deprecated.
    Analyze the relationship of policy measures and parameters of ODE models.
    This analysis will be done at country level because OxCGRT tracks policies at country level.

    Args:
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        tau (int or None): tau value [min]
    """

    @deprecate("PolicyMeasures", version="2.19.1-zeta-fu1")
    def __init__(self, jhu_data, population_data, oxcgrt_data, tau=None):
        # Records
        self.jhu_data = self._ensure_instance(
            jhu_data, JHUData, name="jhu_data")
        # Population
        self.population_data = self._ensure_instance(
            population_data, PopulationData, name="population_data")
        # OxCGRT
        self.oxcgrt_data = self._ensure_instance(
            oxcgrt_data, OxCGRTData, name="oxcgrt_data")
        # tau value must be shared
        self.tau = Validator(tau, "tau").tau(default=None)
        # Init
        self._countries = self._all_countries()
        self._init_scenario()
        self.model = None

    @staticmethod
    def _ensure_instance(target, class_obj, name="target"):
        """
        Ensure the target is an instance of the class object.

        Args:
            target (instance): target to ensure
            parent (class): class object
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        s = f"@{name} must be an instance of {class_obj}, but {type(target)} was applied."
        if not isinstance(target, class_obj):
            raise TypeError(s)
        return target

    def _all_countries(self):
        """
        Return names of countries where records are registered.

        Returns:
            (list[str]): list of country names
        """
        j_list = self.jhu_data.countries(complement=True)
        p_list = self.population_data.countries()
        o_list = self.oxcgrt_data.countries()
        return list(set(j_list) & set(p_list) & set(o_list))

    def _init_scenario(self):
        """
        Initialize the scenario classes of registered countries.
        """
        self.scenario_dict = {
            country: Scenario(
                self.jhu_data, self.population_data, country=country, tau=self.tau)
            for country in self._countries
        }

    def scenario(self, country):
        """
        Return Scenario instance of the country.

        Args:
            country (str): country name

        Raises:
            KeyError: the country is not registered

        Returns:
            covsirphy.Scenario: Scenario instance
        """
        if country not in self.scenario_dict.keys():
            raise KeyError(f"{country} is not registered.")
        return self.scenario_dict[country]

    @property
    def countries(self):
        """
        list[str]: countries to analyze
        """
        return self._countries

    @countries.setter
    def countries(self, country_list):
        selected_set = set(country_list)
        all_set = set(self._all_countries())
        if not selected_set.issubset(all_set):
            un_selectable_set = selected_set - all_set
            un_selectable = ", ".join(list(un_selectable_set))
            raise KeyError(
                f"{un_selectable} cannot be selected because records are not registered.")
        self._countries = country_list

    def trend(self, min_len=2):
        """
        Perform S-R trend analysis for all registered countries.

        Args:
            min_len (int): minimum length of phases to have

        Returns:
            covsirphy.PolicyMeasures: self

        Note:
            Countries which do not have @min_len phases will be un-registered.
        """
        min_len = self._ensure_natural_int(min_len, name="min_len")
        for country in self._countries:
            with contextlib.suppress(ValueError):
                self.scenario_dict[country].trend(set_phases=True, show_figure=False)
        countries = [
            country for country in self._countries
            if len(self.scenario_dict[country][self.MAIN]) >= min_len
        ]
        self.countries = countries
        return self

    def summary(self, columns=None, countries=None):
        """
        Summarize of scenarios.

        Args:
            columns (list[str] or None): columns to show
            countries (list[str] or None): countries to show

        Returns:
            pandas.DataFrame

        Note:
            If @columns is None, all columns will be shown.
        """
        countries = countries or self._countries
        if not isinstance(countries, (list, set)):
            raise TypeError("@countries must be a list or set.")
        dataframes = []
        for country in countries:
            df = self.scenario_dict[country].summary(columns=columns)
            df[self.PHASE] = df.index
            df[self.COUNTRY] = country
            dataframes.append(df)
        summary_df = pd.concat(dataframes, axis=0, ignore_index=True)
        summary_df = summary_df.set_index([self.COUNTRY, self.PHASE])
        return summary_df.fillna(self.NA)

    def phase_len(self):
        """
        Make groups of countries with the length of phases.

        Returns:
            dict(int, list[str]): list of countries with the length of phases
        """
        len_nest = [
            (country, len(self.scenario_dict[country][self.MAIN]))
            for country in self._countries
        ]
        sorted_nest = sorted(len_nest, key=itemgetter(1), reverse=True)
        return {
            length: [country for (country, _) in records]
            for (length, records) in groupby(sorted_nest, key=itemgetter(1))
        }

    def estimate(self, model, n_jobs=-1, **kwargs):
        """
        Estimate the parameter values of phases in the registered countries.

        Args:
            model (covsirphy.ModelBase): ODE model
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()
        """
        model = Validator(model, "model").subclass(ModelBase)
        unit_nest = [
            [
                unit.set_id(
                    country=self.jhu_data.country_to_iso3(country),
                    phase=f"{self.num2str(num):>4}")
                for (num, unit) in enumerate(self.scenario_dict[country][self.MAIN]) if unit]
            for country in self._countries
        ]
        units = self.flatten(unit_nest)
        # Parameter estimation
        mp_estimator = MPEstimator(
            jhu_data=self.jhu_data, population_data=self.population_data,
            model=model, tau=self.tau, **kwargs)
        mp_estimator.add(units)
        results = mp_estimator.run(
            n_jobs=n_jobs, auto_complement=True, **kwargs)
        # Register the results
        for country in self._countries:
            new_units = [
                unit for unit in results
                if unit.id_dict["country"] == self.jhu_data.country_to_iso3(country)]
            self.scenario_dict[country][self.MAIN].replaces(
                phase=None, new_list=new_units, keep_old=True)
        self.model = model
        self.tau = mp_estimator.tau

    @deprecate(
        old="PolicyMeasures.param_history(param: str)",
        new="PolicyMeasures.history(param: str)",
        version="2.9.0")
    def param_history(self, param, roll_window=None, show_figure=True, filename=None, **kwargs):
        return self.history(param, roll_window=None, show_figure=True, filename=None, **kwargs)

    def history(self, param, roll_window=None, show_figure=True, filename=None, **kwargs):
        """
        Return subset of summary and show a figure to show the history of all countries.

        Args:
            param (str): parameter/day parameter/Rt/OxCGRT score to show
            roll_window (int or None): rolling average window if necessary
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of line_plot()

        Returns:
            pandas.DataFrame:
                Index
                    Date (pd.Timestamp) date
                Columns
                    (str) country names
                Values:
                    parameter values
        """
        # Get the parameter value of each date
        df = self.track()
        # Select the param
        if param not in df.columns:
            sel_param_str = ', '.join(df.columns.tolist())
            raise KeyError(
                f"@param must be selected from {sel_param_str}, but {param} was applied.")
        df = df.pivot_table(
            values=param, index=self.DATE, columns=self.COUNTRY, aggfunc="last")
        # Rolling mean
        if roll_window is not None:
            roll_window = self._ensure_natural_int(
                roll_window, name="roll_window")
            df = df.rolling(window=roll_window).mean()
        # Show figure
        if not show_figure:
            return df
        line_plot(
            df, title=f"History of {param} in each country",
            ylabel=param,
            h=1 if param == self.RT else None,
            filename=filename,
            **kwargs
        )
        return df

    @staticmethod
    def _ensure_natural_int(target, name="number", include_zero=False, none_ok=False):
        """
        Ensure a natural (non-negative) number.

        Args:
            target (int or float or str or None): value to ensure
            name (str): argument name of the value
            include_zero (bool): include 0 or not
            none_ok (bool): None value can be applied or not.

        Returns:
            int: as-is the target

        Note:
            When @target is None and @none_ok is True, None will be returned.
            If the value is a natural number and the type was float or string,
            it will be converted to an integer.
        """
        if target is None and none_ok:
            return None
        s = f"@{name} must be a natural number, but {target} was applied"
        try:
            number = int(target)
        except TypeError as e:
            raise TypeError(f"{s} and not converted to integer.") from e
        if number != target:
            raise ValueError(f"{s}. |{target} - {number}| > 0")
        min_value = 0 if include_zero else 1
        if number < min_value:
            raise ValueError(f"{s}. This value is under {min_value}")
        return number

    def track(self):
        """
        Return subset of summary and show a figure to show the history in each country.

        Args:
            param (str): parameter to show
            roll_window (int or None): rolling average window if necessary
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of pd.DataFrame.plot or line_plot()

        Returns:
            pandas.DataFrame: parameter values
                Index
                    reset index
                Columns
                    - Country (str): country name
                    - Date (pd.Timestamp): date
                    - (float): model parameters
                    - (float): model day parameters
                    - Rt (float): reproduction number
                    - (float): OxCGRT values
        """
        if self.model is None:
            raise UnExecutedError("PolicyMeasures.estimate(model)")
        # Get parameter/Rt/data parameter value of each date
        df = self.summary().reset_index().replace(self.NA, None)
        df[self.START] = pd.to_datetime(
            df[self.START], format=self.DATE_FORMAT)
        df[self.END] = pd.to_datetime(df[self.END], format=self.DATE_FORMAT)
        df[self.DATE] = df[[self.START, self.END]].apply(
            lambda x: pd.date_range(x[0], x[1]).tolist(), axis=1)
        df = df.explode(self.DATE)
        cols = [
            self.DATE, self.COUNTRY, *self.model.PARAMETERS, *self.model.DAY_PARAMETERS, self.RT]
        param_df = df.reindex(cols, axis=1)
        # OxCGRT
        oxcgrt_df = self.oxcgrt_data.cleaned()
        sel = oxcgrt_df[self.COUNTRY].isin(self._countries)
        oxcgrt_df = oxcgrt_df.loc[
            sel, [self.DATE, self.COUNTRY, *OxCGRTData.OXCGRT_VARS]]
        # Combine data
        return pd.merge(
            param_df, oxcgrt_df, how="inner", on=[self.COUNTRY, self.DATE])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from autots import AutoTS
from autots.evaluator.auto_ts import fake_regressor
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _AutoTSHandler(Term):
    """Class for forecasting with AutoTS package.

    Args:
        Y (pandas.DataFrame):
            Index
                pandas.Timestamp: Observation date
            Columns
                observed and the target variables (int or float)
        days (int): days to predict
        seed (int or None): random seed
        verbose (int): verbosity
        **kwargs: keyword arguments of autots.AutoTS() except for forecast_length (always the same as @days)

    Note:
        AutoTS package is developed at https://github.com/winedarksea/AutoTS
    """

    def __init__(self, Y, days, seed, verbose, **kwargs):
        self._Y = Validator(Y, "Y").dataframe(time_index=True, empty_ok=False)
        self._days = Validator(days, name="days").int(value_range=(1, None))
        autots_kwargs = {
            "forecast_length": self._days,
            "frequency": "D",
            "no_negatives": True,
            "model_list": "superfast",
            "model_interrupt": True,
            "ensemble": "horizontal",
            "max_generations": 1,
            "num_validations": 2,
            "validation_method": "backwards",
            "random_seed": Validator(seed, name="seed").int(),
            "n_jobs": "auto",
            "verbose": Validator(verbose, name="verbose").int(),
        }
        autots_kwargs.update(kwargs)
        self._autots = AutoTS(**autots_kwargs)
        self._regressor_forecast = None

    def fit(self, X=None):
        """Fit the model with/without other information.

        Args:
            X (pandas.DataFrame or None): information for regression or None (no information)
                Index
                    pandas.Timestamp: Observation date
                Columns
                    observed and the target variables (int or float)

        Returns:
            _AutoTSHandler: self
        """
        if X is not None:
            regressor_train, self._regressor_forecast = fake_regressor(
                self._Y,
                dimensions=4,
                forecast_length=self._days,
                drop_most_recent=self._autots.drop_most_recent,
                aggfunc=self._autots.aggfunc,
                verbose=self._autots.verbose,
            )
        self._autots.fit(self._Y, future_regressor=None if X is None else regressor_train)
        return self

    def predict(self):
        """Return the predicted values with the observed values.

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): Observation date
                Columns
                    as the same as Y of covsirphy._AutoTSHandler()
        """
        return self._autots.predict(future_regressor=self._regressor_forecast).forecast.rename_axis(self.DATE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import pandas as pd
import ruptures as rpt
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.sr_change import _SRChange


class TrendDetector(Term):
    """
    Interface for trend analysis (change point analysis).

    Args:
        data (pandas.DataFrame): data to analyze
            Index:
                reset index
            Column:
                - Date (pd.Timestamp): Observation date
                - Recovered (int): the number of recovered cases
                - Susceptible (int): the number of susceptible cases
        area (str): area name (used in the figure title)
        min_size (int): minimum value of phase length [days], over 2

    Note:
        "Phase" means a sequential dates in which the parameters of SIR-derived models are fixed.
        "Change points" means the dates when trend was changed.
        "Change points" is the same as the start dates of phases except for the 0th phase.
    """

    @deprecate(old="TrendDetector", new="Dynamics", version="2.24.0-xi")
    def __init__(self, data, area="Selected area", min_size=7):
        Validator(data, "data").dataframe(columns=[self.DATE, self.S, self.R])
        # Index: Date, Columns: the number cases
        self._record_df = data.groupby(self.DATE).last()
        # Minimum size of phases
        self._min_size = Validator(min_size, "min_size").int(value_range=(3, None))
        if len(self._record_df) < self._min_size * 2:
            raise ValueError(f"More than {min_size * 2} records must be included because @min_size is {min_size}.")
        # The first/last date
        self._first_point = self._record_df.index.min()
        self._last_point = self._record_df.index.max()
        # Area name
        self._area = area
        # Change points: list[pandas.Timestamp]
        self._points = []

    def reset(self):
        """
        Reset the phase setting with the end dates of the records.

        Returns:
            covsirphy.TrendDetector: self
        """
        self._points = []
        return self

    def dates(self):
        """
        Return the list of start dates and end dates.

        Returns:
            tuple(list[str], list[str]): list of start dates and end dates
        """
        points = self._points[:]
        # Start dates
        start_dates = [date.strftime(self.DATE_FORMAT) for date in [self._first_point, *points]]
        # End dates
        end_dates = [(Validator(date).date() - timedelta(days=1)).strftime(self.DATE_FORMAT) for date in points]
        end_dates.append(self._last_point.strftime(self.DATE_FORMAT))
        return (start_dates, end_dates)

    def summary(self, metric=None, metrics="MSE"):
        """
        Summarize the phases with a dataframe.

        Args:
            metric (str or None): metric name or None (use @metrics)
            metrics (str): alias of @metric

        Returns:
            pandas.Dataframe:
                Index
                    (str): phase names
                Columns
                    - Start (str): star dates
                    - End (str): end dates
                    - Duration (int): phase duration
                    - {metric}_S-R (float): scores on S-R plane with the selected metric

        Note:
            Please refer to covsirphy.Evaluator.score() for metric names
        """
        metric = metric or metrics
        # Phase duration
        start_dates, end_dates = self.dates()
        duration_list = [self.steps(start, end, tau=1440) for (start, end) in zip(start_dates, end_dates)]
        # Scores in S-R plane
        scores = _SRChange(sr_df=self._record_df).score(change_points=self._points, metric=metric)
        return pd.DataFrame(
            {
                self.START: start_dates,
                self.END: end_dates,
                "Duration": duration_list,
                f"{metric}_S-R": scores,
            },
            index=[self.num2str(num) for num in range(len(self._points) + 1)]
        )

    def sr(self, algo="Binseg-normal", **kwargs):
        """
        Perform S-R trend analysis.

        Args:
            algo (str): detection algorithms and models
            kwargs: the other arguments of algorithm classes (ruptures.Pelt, .Binseg, BottomUp)

        Raises:
            UnExpectedValueError: un-expected value was applied as algorithm name

        Returns:
            covsirphy.TrendDetector: self

        Note:
            Candidates of @algo are "Pelt-rbf", "Binseg-rbf", "Binseg-normal", "BottomUp-rbf", "BottomUp-normal".
            Please refer to documentation of ruptures package.
            https://centre-borelli.github.io/ruptures-docs/
        """
        # Set algorithm class
        algo_kwargs = {"jump": 1, "min_size": self._min_size}
        algo_kwargs.update(kwargs)
        algo_dict = {
            "Pelt-rbf": (rpt.Pelt, {"model": "rbf"}),
            "Binseg-rbf": (rpt.Binseg, {"model": "rbf"}),
            "Binseg-normal": (rpt.Binseg, {"model": "normal"}),
            "BottomUp-rbf": (rpt.BottomUp, {"model": "rbf"}),
            "BottomUp-normal": (rpt.BottomUp, {"model": "normal"}),
        }
        Validator([algo], "algo").sequence(candidates=algo_dict.keys())
        algo_kwargs.update(algo_dict[algo][1])
        algorithm = algo_dict[algo][0](
            **Validator(algo_kwargs, "keyword arguments").kwargs(functions=algo_dict[algo][0]))
        # Run trend analysis
        finder = _SRChange(sr_df=self._record_df)
        points = finder.run(algorithm=algorithm, **algo_kwargs)
        self._points = sorted(set(self._points) | set(points))
        return self

    def show(self, **kwargs):
        """
        Show the trend on S-R plane.

        Args:
            kwargs: keyword arguments of covsirphy.trend_plot()
        """
        finder = _SRChange(sr_df=self._record_df)
        finder.show(self._points, self._area, **kwargs)


class Trend(TrendDetector):
    """
    Deprecated. Please use TrendDetector class.
    """
    @deprecate("covsirphy.Trend", new="covsirphy.TrendDetector", version="2.17.0-zeta")
    def __init__(self, data, area="Selected area", min_size=5):
        super().__init__(data, area=area, min_size=min_size)


class ChangeFinder(TrendDetector):
    """
    Deprecated. Please use TrendDetector class.
    """
    @deprecate("covsirphy.ChangeFinder", new="covsirphy.TrendDetector", version="2.17.0-zeta")
    def __init__(self, data, area="Selected area", min_size=5):
        super().__init__(data, area=area, min_size=min_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.term import Term
from covsirphy._deprecated.db_base import _RemoteDatabase


class _OWID(_RemoteDatabase):
    """
    Access "Our World In Data".
    https://github.com/owid/covid-19-data/tree/master/public/data
    https://ourworldindata.org/coronavirus

    Args:
        filename (str): CSV filename to save records
        iso3 (str or None): ignored
    """
    # URL for vaccine data
    URL_V = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/vaccinations/"
    URL_V_REC = f"{URL_V}vaccinations.csv"
    URL_V_LOC = f"{URL_V}locations.csv"
    # URL for PCR data
    URL_P = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/testing/"
    URL_P_REC = f"{URL_P}covid-testing-all-observations.csv"
    # Citation
    CITATION = "Hasell, J., Mathieu, E., Beltekian, D. et al." \
        " A cross-country database of COVID-19 testing. Sci Data 7, 345 (2020)." \
        " https://doi.org/10.1038/s41597-020-00688-8"
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    COL_DICT = {
        "date": Term.DATE,
        "iso_code": Term.ISO3,
        Term.PROVINCE: Term.PROVINCE,
        "vaccines": Term.PRODUCT,
        "total_vaccinations": Term.VAC,
        "total_boosters": Term.VAC_BOOSTERS,
        "people_vaccinated": Term.V_ONCE,
        "people_fully_vaccinated": Term.V_FULL,
        "Cumulative total": Term.TESTS,
    }

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()

        Note:
            If @verbose is equal to or over 1, how to show the list will be explained.
        """
        # Download datasets
        if verbose:
            print("Retrieving datasets from Our World In Data https://github.com/owid/covid-19-data/")
        # Vaccinations
        v_rec_cols = [
            "date", "iso_code", "total_vaccinations", "total_boosters", "people_vaccinated", "people_fully_vaccinated"]
        v_rec_df = pd.read_csv(self.URL_V_REC, usecols=v_rec_cols)
        v_loc_df = pd.read_csv(self.URL_V_LOC, usecols=["iso_code", "vaccines"])
        v_df = v_rec_df.merge(v_loc_df, how="left", on="iso_code")
        # Tests
        pcr_rec_cols = ["ISO code", "Date", "Daily change in cumulative total", "Cumulative total"]
        pcr_df = pd.read_csv(self.URL_P_REC, usecols=pcr_rec_cols)
        pcr_df = pcr_df.rename(columns={"ISO code": "iso_code", "Date": "date"})
        pcr_df["cumsum"] = pcr_df.groupby("iso_code")["Daily change in cumulative total"].cumsum()
        pcr_df = pcr_df.assign(tests=lambda x: x["Cumulative total"].fillna(x["cumsum"]))
        # Combine data (vaccinations/tests)
        df = v_df.merge(pcr_df, how="outer", on=["iso_code", "date"])
        # Location (iso3/province)
        df = df.loc[~df["iso_code"].str.contains("OWID_")]
        df = df.loc[~df["iso_code"].isna()]
        df[self.PROVINCE] = self.NA
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from covsirphy.util.error import deprecate
from covsirphy.util.term import Term
from covsirphy._deprecated.dataloader import DataLoader


class COVID19DataHub(Term):
    """
    Deprecated. Please use DataLoader() class.
    Load datasets retrieved from COVID-19 Data Hub.
    https://covid19datahub.io/

    Args:
        filename (str): CSV filename to save records
    """

    @deprecate("covsirphy.COVID19DataHub()", new="covsirphy.DataLoader()", version="2.21.0-eta")
    def __init__(self, filename):
        self._filepath = Path(filename)

    def load(self, name="jhu", force=True, verbose=1):
        """
        Load the datasets of COVID-19 Data Hub and create dataset object.

        Args:
            name (str): name of dataset, "jhu", "population", "oxcgrt" or "pcr"
            force (bool): if True, always download the dataset from the server
            verbose (int): level of verbosity

        Returns:
            covsirphy.CleaningBase: the dataset

        Note:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation etc.
        """
        loader = DataLoader(
            directory=self._filepath.parent,
            update_interval=0 if force else 12,
            basename_dict={"covid19dh": self._filepath.name},
            verbose=verbose
        )
        if force:
            self._filepath.unlink()
        return {
            "jhu": loader.jhu,
            "population": loader.population,
            "oxcgrt": loader.oxcgrt,
            "pcr": loader.pcr
        }[name]()

    @property
    def primary(self):
        """
        str: the list of primary sources.
        """
        loader = DataLoader(directory=self._filepath.parent)
        return loader.covid19dh_citation



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from datetime import datetime
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class VaccineData(CleaningBase):
    """
    Data cleaning of vaccination dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - Country: country/region name
                - ISO3: ISO 3166-1 alpha-3, like JPN
                - Province: province/prefecture/state name
                - Product: vaccine product names
                - Vaccinations: cumulative number of vaccinations
                - Vaccinations: cumulative number of booster vaccinations
                - Vaccinated_once: cumulative number of people who received at least one vaccine dose
                - Vaccinated_full: cumulative number of people who received all doses prescrived by the protocol
        citation (str or None): citation or None (empty)
        kwargs: the other arguments will be ignored

    Note:
        Either @filename (high priority) or @data must be specified.

    Note:
        Columns of VaccineData.cleaned():
            - Date (pandas.TimeStamp): observation dates
            - Country (pandas.Category): country (or province) names
            - ISO3 (pandas.Category): ISO3 codes
            - Province (pandas.Category): province/prefecture/state name
            - Product (pandas.Category): vaccine product names
            - Vaccinations (int): cumulative number of vaccinations
            - Vaccinations (int): cumulative number of booster vaccinations
            - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
            - Vaccinated_full (int): cumulative number of people who received all doses prescribed by the protocol
    """

    @deprecate(old="VaccineData", new="DataEngineer", version="2.24.0-xi")
    def __init__(self, filename=None, data=None, citation=None, **kwargs):
        self._subset_cols = [self.DATE, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL]
        self._raw_cols = [
            self.DATE, self.ISO3, self.COUNTRY, self.PROVINCE, self.PRODUCT,
            self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL]
        # Raw data
        self._raw = self._parse_raw(filename, data, self._raw_cols)
        # Backward compatibility
        if self._raw.empty:
            self._raw = self._retrieve(filename, **kwargs)
        if self.PROVINCE not in self._raw:
            self._raw[self.PROVINCE] = self.NA
        # Data cleaning
        self._cleaned_df = pd.DataFrame(columns=self._raw_cols) if self._raw.empty else self._cleaning()
        # Citation
        self._citation = citation or ""
        # Directory that save the file
        if filename is None:
            self._dirpath = Path("input")
        else:
            Path(filename).parent.mkdir(exist_ok=True, parents=True)
            self._dirpath = Path(filename).resolve().parent

    @deprecate(
        "vaccine_data = cs.VaccineData()",
        new="vaccine_data = cs.DataLoader().vaccine()", version="2.21.0-iota-fu1")
    def _retrieve(self, filename, verbose=1, **kwargs):
        """
        Retrieve the dataset from server.
        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity
        Returns:
            pd.DataFrame:
                Index reset index
                Columns Date, Country, Product, Vaccinations
        """
        URL = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/vaccinations/"
        URL_REC = f"{URL}vaccinations.csv"
        URL_LOC = f"{URL}locations.csv"
        # Show URL
        if verbose:
            print("Retrieving COVID-19 vaccination dataset from https://github.com/owid/covid-19-data/")
        # Download datasets and merge them
        rename_dict = {
            "date": self.DATE, "location": self.COUNTRY, "iso_code": self.ISO3,
            "vaccines": self.PRODUCT, "total_vaccinations": self.VAC, "total_boosters": self.VAC_BOOSTERS,
            "people_vaccinated": self.V_ONCE,
            "people_fully_vaccinated": self.V_FULL,
        }
        rec_df = pd.read_csv(URL_REC, usecols=list(set(rename_dict) - {"vaccines"}))
        loc_df = pd.read_csv(URL_LOC, usecols=["location", "vaccines"])
        df = rec_df.merge(loc_df, how="left", on="location")
        df = df.rename(rename_dict, axis=1)
        # Save the dataframe as CSV file
        df.to_csv(filename, index=False)
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                - Date (pandas.TimeStamp): observation dates
                - Country (pandas.Category): country (or province) names
                - ISO3 (pandas.Category): ISO3 codes
                - Province: province/prefecture/state name
                - Product (pandas.Category): vaccine product names
                - Vaccinations (int): cumulative number of vaccinations
                - Vaccinations (int): cumulative number of booster vaccinations
                - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._raw.copy()
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Fill in NA values
        for col in [self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL]:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        country_df = df.loc[:, [self.COUNTRY, self.ISO3, self.PRODUCT]].drop_duplicates()
        # Extent dates to today
        today_date = datetime.now().replace(
            hour=00, minute=00, second=00, microsecond=00
        )

        df = df.pivot_table(
            values=[self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL],
            index=self.DATE, columns=[self.COUNTRY, self.PROVINCE], aggfunc="last")
        df = df.reindex(pd.date_range(df.index[0], today_date, freq="D"))
        df.index.name = self.DATE
        df = df.ffill().fillna(0).astype(np.int64).stack().stack().reset_index()
        df.sort_values(by=[self.COUNTRY, self.PROVINCE, self.DATE], ignore_index=True, inplace=True)
        df = df.merge(country_df, on=self.COUNTRY)
        # Set dtype for category data
        for col in [self.COUNTRY, self.ISO3, self.PROVINCE, self.PRODUCT]:
            df[col] = df[col].astype("category")
        return df.loc[:, self._raw_cols]

    def subset(self, country, province=None, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            product (str or None): vaccine product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinations_boosters (int): the number of booster vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._cleaned_df.copy()
        # Subset by country
        iso3 = self.ensure_country_name(country)
        df = df.loc[(df[self.ISO3] == iso3) & (df[self.PROVINCE] == (province or self.NA))]
        # Subset by product name
        if product is not None:
            df = df.loc[df[self.PRODUCT] == product]
        # Subset with start date
        if start_date is not None:
            df = df.loc[df[self.DATE] >= self._ensure_date(start_date)]
        # Subset with end date
        if end_date is not None:
            df = df.loc[df[self.DATE] <= self._ensure_date(end_date)]
        # Check records were found
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=iso3, province=product, start_date=start_date, end_date=end_date)
        return df.loc[:, self._subset_cols].reset_index(drop=True)

    def records(self, country, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str or None): country name or ISO3 code
            product (str or None): product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinations (int): the number of booster vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        return self.subset(
            country=country, product=product, start_date=start_date, end_date=end_date)

    def total(self):
        """
        Calculate total values of the cleaned dataset.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinations_boosters (int): the number of booster vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._cleaned_df.copy()
        # Select 'World' data
        df = df.loc[df[self.COUNTRY] == "World"]
        # Resampling
        df = df.set_index(self.DATE).resample("D").sum()
        return df.reset_index()[self._subset_cols]

    def map(self, country=None, variable="Vaccinations", date=None, **kwargs):
        """
        Create colored map with the number of vaccinations.

        Args:
            country (None): always None
            variable (str): variable to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @country was specified
        """
        if country is not None:
            raise NotImplementedError("@country cannot be specified, always None.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        title = f'Global: the number of {variable.lower()} on {date_str}'
        # Global map
        return self._colored_map_global(variable=variable, title=title, date=date, **kwargs)
